"""Module for converting field maps of ADNI."""
import json
import os
import re
import shutil
from enum import Enum
from os import PathLike
from pathlib import Path
from typing import Callable, Iterable, List, Optional

import pandas as pd

from clinica.iotools.converters.adni_to_bids.adni_utils import load_clinical_csv
from clinica.utils.stream import cprint

__all__ = ["convert_fmap"]


def convert_fmap(
    source_dir: PathLike,
    csv_dir: PathLike,
    destination_dir: PathLike,
    conversion_dir: PathLike,
    subjects: Iterable[str],
    mod_to_update: bool = False,
    n_procs: int = 1,
):
    """Convert field map images of ADNI into BIDS format.

    Parameters
    ----------
    source_dir : PathLike
        Path to the ADNI directory.

    csv_dir : PathLike
        Path to the clinical data directory.

    destination_dir : PathLike
        Path to the destination BIDS directory.

    conversion_dir : PathLike
        Path to the TSV files including the paths to original images.

    subjects : List of str, optional
        List of subjects.

    mod_to_update : bool
        If True, pre-existing images in the BIDS directory
        will be erased and extracted again.
    """

    from os import path

    import pandas as pd

    from clinica.iotools.converters.adni_to_bids.adni_utils import (
        ADNIModalityConverter,
        paths_to_bids,
    )
    from clinica.utils.stream import cprint

    csv_dir = Path(csv_dir)
    source_dir = Path(source_dir)
    conversion_dir = Path(conversion_dir)

    cprint(
        f"Calculating paths of fMRI field maps (FMAPs). Output will be stored in {conversion_dir}.",
        lvl="debug",
    )

    images = compute_fmap_path(source_dir, csv_dir, subjects, conversion_dir)

    cprint("Paths of field maps found. Exporting images into BIDS ...")

    paths_to_bids(
        images, destination_dir, ADNIModalityConverter.FMAP, mod_to_update=mod_to_update
    )
    reorganize_fmaps(destination_dir)

    cprint(msg="Field maps conversion done.", lvl="debug")


def compute_fmap_path(
    source_dir: Path, csv_dir: Path, subjs_list: Iterable[str], conversion_dir: Path
) -> pd.DataFrame:
    """Compute the paths to fMR images.

    Args:
        source_dir: path to the ADNI directory
        csv_dir: path to the clinical data directory
        subjs_list: subjects list
        conversion_dir: path to the TSV files including the paths to original images

    Returns:
        pandas Dataframe containing the path for each fmri
    """
    from os import path
    from pathlib import Path

    import pandas as pd

    from ._image_path_utils import find_image_path
    from ._visits_utils import visits_to_timepoints

    fmap_col = [
        "Subject_ID",
        "VISCODE",
        "Visit",
        "Sequence",
        "Scan_Date",
        "Study_ID",
        "Field_Strength",
        "Series_ID",
        "Image_ID",
    ]
    fmap_df = pd.DataFrame(columns=fmap_col)
    fmap_dfs_list = []
    adni_merge = load_clinical_csv(csv_dir, "ADNIMERGE")

    mayo_mri_qc = load_clinical_csv(csv_dir, "MAYOADIRL_MRI_IMAGEQC_12_08_15")
    mayo_mri_qc = mayo_mri_qc[mayo_mri_qc.series_type == "fMRI"]
    mayo_mri_qc.columns = [x.upper() for x in mayo_mri_qc.columns]

    mayo_mri_qc3 = load_clinical_csv(csv_dir, "MAYOADIRL_MRI_QUALITY_ADNI3")
    mayo_mri_qc3 = mayo_mri_qc3[mayo_mri_qc3.SERIES_TYPE == "EPB"]

    # Concatenating visits in both QC files
    mayo_mri_qc = pd.concat(
        [mayo_mri_qc, mayo_mri_qc3], axis=0, ignore_index=True, sort=False
    )

    mri_list = load_clinical_csv(csv_dir, "MRILIST")

    # Selecting only fMRI images that are not Multiband
    mri_list = mri_list[
        mri_list.SEQUENCE.str.contains("apping")
    ]  # 'apping' includes all field map scans, but not others

    for subj in subjs_list:
        # Filter ADNIMERGE, MRI_LIST and QC for only one subject and sort the rows/visits by examination date
        adnimerge_subj = adni_merge[adni_merge.PTID == subj]
        adnimerge_subj = adnimerge_subj.sort_values("EXAMDATE")

        mri_list_subj = mri_list[mri_list.SUBJECT == subj]
        mri_list_subj = mri_list_subj.sort_values("SCANDATE")

        mayo_mri_qc_subj = mayo_mri_qc[mayo_mri_qc.RID == int(subj[-4:])]

        # Obtain corresponding timepoints for the subject visits
        visits = visits_to_timepoints(subj, mri_list_subj, adnimerge_subj, "FMAP")

        for visit_info, visit_str in visits.items():
            timepoint = visit_info[0]

            visit_mri_list = mri_list_subj[mri_list_subj.VISIT == visit_str]

            image = fmap_image(
                subj, timepoint, visits[visit_info], visit_mri_list, mayo_mri_qc_subj
            )

            if image is not None:
                row_to_append = pd.DataFrame(
                    image,
                    index=[
                        "i",
                    ],
                )
                fmap_dfs_list.append(row_to_append)

        if fmap_dfs_list:
            fmap_df = pd.concat(fmap_dfs_list, ignore_index=True)

    # Exceptions
    # ==========
    conversion_errors = [
        # Multiple images
        ("029_S_2395", "m72"),
        # Real/Imaginary
        ("002_S_1261", "m60"),
        ("002_S_1261", "m72"),
        ("002_S_1261", "m84"),
        ("002_S_1261", "m96"),
        ("006_S_4485", "bl"),
        ("006_S_4485", "m03"),
        ("006_S_4485", "m06"),
        ("006_S_4485", "m12"),
        ("006_S_4485", "m24"),
        ("006_S_4485", "m48"),
        # Unrecognized BIDSCase
        ("006_S_4485", "m78"),
        ("009_S_4388", "m03"),
        ("009_S_4388", "m06"),
        ("009_S_4388", "m12"),
        ("009_S_4388", "m24"),
        ("009_S_4388", "m48"),
        ("023_S_4115", "bl"),
        ("023_S_4115", "m03"),
        ("023_S_4115", "m06"),
        ("023_S_4115", "m12"),
        ("023_S_4115", "m24"),
        ("023_S_4115", "m48"),
        ("123_S_4127", "bl"),
        ("123_S_4127", "m12"),
        ("123_S_4127", "m24"),
        ("123_S_4127", "m36"),
        # Missing EchoTime Keys
        ("006_S_4485", "m90"),
        ("036_S_6088", "m12"),
        ("123_S_4127", "m84"),
        # Missing DICOMS
        ("023_S_4115", "m126"),
        ("177_S_6448", "m24"),
    ]

    # Removing known exceptions from images to convert
    if not fmap_df.empty:
        error_ind = fmap_df.index[
            fmap_df.apply(
                lambda x: ((x.Subject_ID, x.VISCODE) in conversion_errors), axis=1
            )
        ]
        fmap_df.drop(error_ind, inplace=True)

    # Checking for images paths in filesystem
    images = find_image_path(fmap_df, source_dir, "FMAP")
    images.to_csv(conversion_dir / "fmap_paths.tsv", sep="\t", index=False)

    return images


def fmap_image(
    subject_id: str,
    timepoint: str,
    visit_str: str,
    visit_mri_list: list[pd.DataFrame],
    mri_qc_subj: pd.DataFrame,
) -> Optional[dict]:
    """
    One image among those in the input list is chosen according to QC
    and then corresponding metadata is extracted to a dictionary.

    Args:
        subject_id: Subject identifier
        timepoint: Visit code
        visit_str: Visit name
        visit_mri_list: List of images metadata
        mri_qc_subj: Dataframe containing list of QC of scans for the subject

    Returns: dictionary - contains image metadata
             None - no image was able to be selected from the MRI list based on IDs provided
    """
    from clinica.iotools.converter_utils import replace_sequence_chars

    from ._qc_utils import select_image_qc

    mri_qc_subj.columns = [x.lower() for x in mri_qc_subj.columns]
    sel_image = select_image_qc(list(visit_mri_list.IMAGEUID), mri_qc_subj)

    if not sel_image:
        return None

    sel_scan = visit_mri_list[visit_mri_list.IMAGEUID == sel_image].iloc[0]

    image_dict = {
        "Subject_ID": subject_id,
        "VISCODE": timepoint,
        "Visit": visit_str,
        "Sequence": replace_sequence_chars(sel_scan.SEQUENCE),
        "Scan_Date": sel_scan["SCANDATE"],
        "Study_ID": str(int(sel_scan.STUDYID)),
        "Series_ID": str(int(sel_scan.SERIESID)),
        "Image_ID": str(int(sel_scan.IMAGEUID)),
        "Field_Strength": sel_scan.MAGSTRENGTH,
    }

    return image_dict


class BIDSFMAPCase(str, Enum):
    ALREADY_RENAMED = "already_renamed"
    EMPTY_FOLDER = "empty_folder"
    ONE_PHASE_TWO_MAGNITUDES = "one_phase_two_magnitudes"
    TWO_PHASES_TWO_MAGNITUDES = "two_phases_two_magnitudes"
    DIRECT_FIELDMAPS = "direct_fieldmaps"
    NOT_SUPPORTED = "not_supported"


def phase_magnitude_renamer(old_extension: str, case: BIDSFMAPCase) -> str:
    # Assuming case 1 with 1 or 2 magnitudes have the same outputs from dcm2nix
    if (
        case != BIDSFMAPCase.ONE_PHASE_TWO_MAGNITUDES
        and case != BIDSFMAPCase.TWO_PHASES_TWO_MAGNITUDES
    ):
        raise NotImplementedError(
            f"No renaming should be performed for case {case.value}"
        )

    magnitude_match = re.match(pattern="^e([1-2])$", string=old_extension)
    if magnitude_match:
        return f"magnitude{magnitude_match.group(1)}"

    phase_match = re.match(pattern="^e([1-2])_ph$", string=old_extension)
    if phase_match:
        if case == BIDSFMAPCase.ONE_PHASE_TWO_MAGNITUDES:
            return "phasediff"
        elif case == BIDSFMAPCase.TWO_PHASES_TWO_MAGNITUDES:
            return f"phase{phase_match.group(1)}"

    raise ValueError(f"Extension {old_extension} not taken in charge.")


def rename_files(fmap_path: Path, case: BIDSFMAPCase):
    filenames = [f.name for f in fmap_path.iterdir() if not f.name.startswith(".")]
    pattern = r"(.*_)fmap_(.*?)(\..*)$"
    for previous_filename in filenames:
        rgx = re.search(pattern, previous_filename)
        if rgx:
            cut, extension, type = rgx.group(1), rgx.group(2), rgx.group(3)
            new_name = f"{cut}{phase_magnitude_renamer(extension, case)}{type}"
            os.rename(fmap_path / previous_filename, fmap_path / new_name)
        else:
            raise ValueError(f"Invalid file {previous_filename} was found.")


def fmap_case_handler_factory(case: BIDSFMAPCase) -> Callable:
    if case == BIDSFMAPCase.ALREADY_RENAMED:
        return already_renamed_handler
    if case == BIDSFMAPCase.EMPTY_FOLDER:
        return empty_folder_handler
    if case == BIDSFMAPCase.ONE_PHASE_TWO_MAGNITUDES:
        return one_phase_two_magnitudes_handler
    if case == BIDSFMAPCase.TWO_PHASES_TWO_MAGNITUDES:
        return two_phases_two_magnitudes_handler
    if case == BIDSFMAPCase.DIRECT_FIELDMAPS:
        return direct_fieldmaps_handler
    if case == BIDSFMAPCase.NOT_SUPPORTED:
        return not_supported_handler


def get_json_file_matching_pattern(fmap_path: Path, pattern: str) -> Path:
    json_file = [
        f
        for f in fmap_path.iterdir()
        if not f.name.startswith(".") and f.name.endswith(pattern)
    ]
    if len(json_file) != 1:
        msg = f"Expected only a single JSON file ending in '{pattern}' in {fmap_path}, but got:"
        msg += "\n".join([f.name for f in json_file])
        raise ValueError(msg)
    return json_file[0]


def check_json_contains_keys(json_file: Path, keys: Iterable[str]) -> bool:
    with open(json_file, "r") as file:
        json_data = json.load(file)
    missing_keys = [key for key in keys if key not in json_data]
    if missing_keys:
        cprint(
            f"Invalid file {json_file}, missing the following keys: {missing_keys}.",
            lvl="warning",
        )
        return False
    return True


def already_renamed_handler(fmap_path: Path):
    cprint(
        f"Files for subject {fmap_path.parent.parent.name},"
        f"session {fmap_path.parent.name} were already renamed.",
        lvl="info",
    )


def empty_folder_handler(fmap_path: Path):
    cprint(
        f"Folder for subject {fmap_path.parent.parent.name},"
        f"session {fmap_path.parent.name} is empty",
        lvl="warning",
    )


def one_phase_two_magnitudes_handler(fmap_path: Path):
    """Performs the checks and renaming associated to BIDS specifications, case 1 for fieldmaps"""
    cprint(
        f"BIDS Case 1 : expecting 1 phasediff and at least 1 magnitude files for subject"
        f"{fmap_path.parent.parent.name}, session {fmap_path.parent.name}.",
        lvl="info",
    )
    json_file = get_json_file_matching_pattern(fmap_path, pattern="ph.json")
    if check_json_contains_keys(json_file, ("EchoTime1", "EchoTime2")):
        rename_files(fmap_path, BIDSFMAPCase.ONE_PHASE_TWO_MAGNITUDES)
    else:
        not_supported_handler(fmap_path)


def two_phases_two_magnitudes_handler(fmap_path: Path):
    """Performs the checks and renaming for BIDS spec case 2 for fieldmaps"""
    cprint(
        f"BIDS Case 2 : expecting 2 phase and 2 magnitude files for subject {fmap_path.parent.parent.name},"
        f"session {fmap_path.parent.name}.",
        lvl="info",
    )
    json_files = [
        f
        for f in fmap_path.iterdir()
        if not f.name.startswith(".") and f.name.endswith("ph.json")
    ]
    if all([check_json_contains_keys(f, ("EchoTime",)) for f in json_files]):
        rename_files(fmap_path, BIDSFMAPCase.TWO_PHASES_TWO_MAGNITUDES)
    else:
        not_supported_handler(fmap_path)


def direct_fieldmaps_handler(fmap_path: Path):
    cprint(
        f"Assuming BIDS Case 3 for subject {fmap_path.parent.parent.name},"
        f"session {fmap_path.parent.name}. Not supported yet.",
        lvl="info",
    )
    # Assuming filename ends with "fmap" for the fieldmap
    json_file = get_json_file_matching_pattern(fmap_path, pattern="fmap.json")
    check_json_contains_keys(json_file, ("Units",))
    not_supported_handler(fmap_path)


def not_supported_handler(fmap_path: Path):
    to_delete = os.listdir(fmap_path)
    cprint(
        f"No currently supported BIDS case was recognized for subject {fmap_path.parent.parent.name},"
        f"session {fmap_path.parent.name}."
        f"The following files will be deleted : {to_delete}.",
        lvl="warning",
    )
    for file in to_delete:
        os.remove(fmap_path / file)


def infer_case_fmap(fmap_path: Path) -> BIDSFMAPCase:
    files = [f.name for f in fmap_path.iterdir() if not f.name.startswith(".")]
    nb_files = len(files)

    extensions = set(
        re.search(r"fmap(.*).json", f).group(1)
        for f in files
        if re.search(r"fmap(.*).json", f)
    )

    if nb_files == 0:
        return BIDSFMAPCase.EMPTY_FOLDER
    elif all([re.search(r"magnitude|phase", f) for f in files]):
        return BIDSFMAPCase.ALREADY_RENAMED
    elif nb_files == 4:
        if extensions == {""}:
            return BIDSFMAPCase.DIRECT_FIELDMAPS
        else:
            return BIDSFMAPCase.ONE_PHASE_TWO_MAGNITUDES
    elif nb_files == 6 and extensions == {"_e1", "_e2", "_e2_ph"}:
        return BIDSFMAPCase.ONE_PHASE_TWO_MAGNITUDES
    elif nb_files == 8 and extensions == {"_e1", "_e2", "_e1_ph", "_e2_ph"}:
        return BIDSFMAPCase.TWO_PHASES_TWO_MAGNITUDES
    else:
        return BIDSFMAPCase.NOT_SUPPORTED


def reorganize_fmaps(bids_path: Path):
    for subject in (
        folder for folder in bids_path.iterdir() if folder.name.startswith("sub-ADNI")
    ):
        for session in (
            folder
            for folder in subject.iterdir()
            if folder.name.startswith("ses-") and folder.is_dir()
        ):
            fmap_path = bids_path / subject / session / "fmap"
            if fmap_path.exists():
                fmap_case_handler_factory(infer_case_fmap(fmap_path))(fmap_path)



import pydra
from nipype.algorithms.misc import Gunzip
from pydra import Workflow
from pydra.tasks.nipype1.utils import Nipype1Task

from clinica.pydra.engine import clinica_io


def _check_pipeline_parameters(parameters: dict) -> dict:
    """Check the parameters passed to the pipeline.

    Parameters
    ----------
    parameters : dict
        Dictionary of parameters to analyze.

    Returns
    -------
    dict :
        Cleaned dictionary of parameters.
    """
    from clinica.utils.exceptions import ClinicaException

    # PET-Volume pipeline
    if parameters["orig_input_data_volume"] == "pet-volume":
        for param in ("acq_label", "suvr_reference_region"):
            if not parameters[param]:
                raise ClinicaException(
                    f"You selected pet-volume pipeline without setting --{param} flag. "
                    "Clinica will now exit."
                )

    # Custom pipeline
    if parameters["orig_input_data_volume"] == "custom-pipeline":
        if not all([parameters["custom_file"], parameters["measure_label"]]):
            raise ClinicaException(
                "You must set --measure_label and --custom_file flags."
            )
        if not parameters["custom_file"]:
            raise ClinicaException(
                "Custom pipeline was selected but no 'custom_file' was specified."
            )
    return parameters


def _build_query(parameters: dict) -> dict:
    input_name = parameters["orig_input_data_volume"].replace("-", "_")
    query = {
        "group_label": parameters["group_label_dartel"],
        "fwhm": parameters["full_width_at_half_maximum"],
    }
    if input_name == "pet_volume":
        query.update(
            {
                "acq_label": parameters["acq_label"],
                "suvr_reference_region": parameters["suvr_reference_region"],
                "use_brainmasked_image": True,
                "use_pvc_data": parameters["use_pvc_data"],
            }
        )
    elif input_name == "t1_volume":
        query.update({"tissue_number": 1, "modulation": True})

    elif input_name == "custom_pipeline":
        query = {
            "pattern": parameters["custom_file"],
            "description": "custom file provided by user",
        }
    return query


@clinica_io
def build_core_workflow(name: str = "core", parameters={}) -> Workflow:
    """Build the core workflow for the Statistics Volume pipeline.

    Parameters
    ----------
    name : str, optional
        The name of the workflow. Default="core".

    parameters : dict, optional
        Dictionary of parameters to be used
        within the workflow.
        Default={}.

    Returns
    -------
    wf : Workflow
        The core workflow.
    """
    from os.path import join
    from typing import Any

    import clinica.pydra.statistics_volume.task as utils
    from clinica.utils.filemanip import get_parent
    from clinica.utils.spm import use_spm_standalone_if_available

    use_spm_standalone_if_available()

    parameters = _check_pipeline_parameters(parameters)

    input_name = parameters["orig_input_data_volume"].replace("-", "_")
    query = _build_query(parameters)

    input_spec = pydra.specs.SpecInfo(
        name="Input",
        fields=[
            ("_graph_checksums", Any),
            (input_name, dict, query, {"mandatory": True}),
        ],
        bases=(pydra.specs.BaseSpec,),
    )
    wf = Workflow(name, input_spec=input_spec)
    wf.add(
        Nipype1Task(
            name=f"unzip_nii",
            interface=Gunzip(),
            in_file=getattr(wf.lzin, input_name),
        )
        .split("in_file")
        .combine("in_file")
    )
    wf.add(
        utils.get_group_1_and_2_task(
            name="get_groups",
            tsv=parameters["tsv_file"],
            contrast=parameters["contrast"],
        )
    )
    # 1. Model creation
    # We use overwrite option to be sure this node is always run so that it can delete the output dir if it
    # already exists (this may cause error in output files otherwise)
    wf.add(
        utils.write_matlab_model_task(
            name="model_creation",
            tsv=parameters["tsv_file"],
            contrast=parameters["contrast"],
            template_file=join(
                get_parent(__file__, n=3),
                "pipelines",
                "statistics_volume",
                "template_model_creation.m",
            ),
            file_list=wf.unzip_nii.lzout.out_file,
            idx_group1=wf.get_groups.lzout.idx_group1,
            idx_group2=wf.get_groups.lzout.idx_group2,
        )
    )
    wf.add(
        utils.run_m_script_task(
            name="run_spm_model_creation",
            m_file=wf.model_creation.lzout.current_model,
        )
    )
    # 2. Model estimation
    wf.add(
        utils.clean_template_file_task(
            name="model_estimation",
            mat_file=wf.run_spm_model_creation.lzout.output_mat_file,
            template_file=join(
                get_parent(__file__, n=3),
                "pipelines",
                "statistics_volume",
                "template_model_estimation.m",
            ),
        )
    )
    wf.add(
        utils.run_m_script_task(
            name="run_spm_model_estimation",
            m_file=wf.model_estimation.lzout.current_model_estimation,
        )
    )
    # 3. Contrast
    wf.add(
        utils.clean_spm_contrast_file_task(
            name="model_contrast",
            mat_file=wf.run_spm_model_estimation.lzout.output_mat_file,
            template_file=join(
                get_parent(__file__, n=3),
                "pipelines",
                "statistics_volume",
                "template_model_contrast.m",
            ),
            covariates=wf.model_creation.lzout.covariates,
            class_names=wf.get_groups.lzout.class_names,
        )
    )
    wf.add(
        utils.run_m_script_task(
            name="run_spm_model_contrast",
            m_file=wf.model_contrast.lzout.current_model_estimation,
        )
    )
    # 4. Results
    wf.add(
        utils.clean_spm_result_file_task(
            name="model_result_no_correction",
            mat_file=wf.run_spm_model_contrast.lzout.output_mat_file,
            template_file=join(
                get_parent(__file__, n=3),
                "pipelines",
                "statistics_volume",
                "template_model_results.m",
            ),
            method="none",
            threshold=parameters["cluster_threshold"],
        )
    )
    wf.add(
        utils.run_m_script_task(
            name="run_spm_model_result_no_correction",
            m_file=wf.model_result_no_correction.lzout.current_model_result,
        )
    )
    wf.add(
        utils.copy_and_rename_spm_output_files_task(
            name="read_output_node",
            spm_mat=wf.run_spm_model_result_no_correction.lzout.output_mat_file,
            class_names=wf.get_groups.lzout.class_names,
            covariates=wf.model_creation.lzout.covariates,
            group_label=parameters["group_label"],
            fwhm=parameters["full_width_at_half_maximum"],
            measure=parameters["measure_label"],
        )
    )
    wf.set_output(
        [
            ("spmT_0001", wf.read_output_node.lzout.spmT_0001),
            ("spmT_0002", wf.read_output_node.lzout.spmT_0002),
            ("spm_figures", wf.read_output_node.lzout.spm_figures),
            ("variance_of_error", wf.read_output_node.lzout.variance_of_error),
            ("resels_per_voxels", wf.read_output_node.lzout.resels_per_voxels),
            ("mask", wf.read_output_node.lzout.mask),
            ("regression_coeff", wf.read_output_node.lzout.regression_coeff),
            ("contrasts", wf.read_output_node.lzout.contrasts),
        ]
    )
    return wf



from typing import Optional

import click

from clinica import option
from clinica.iotools.converters import cli_param


@click.command(name="aibl-to-bids")
@cli_param.dataset_directory
@cli_param.clinical_data_directory
@cli_param.bids_directory
@cli_param.clinical_data_only
@click.option(
    "--overwrite",
    is_flag=True,
    help="Overwrites previously written nifti and json files.",
)
@option.global_option_group
@option.n_procs
def cli(
    dataset_directory: str,
    clinical_data_directory: str,
    bids_directory: str,
    clinical_data_only: bool = False,
    overwrite: bool = False,
    n_procs: Optional[int] = None,
) -> None:
    """AIBL to BIDS converter.

    Convert the imaging and clinical data of AIBL (https://aibl.csiro.au/adni/index.html), located in DATASET_DIRECTORY
    and CLINICAL_DATA_DIRECTORY respectively, to a BIDS dataset in the target BIDS_DIRECTORY.
    """
    from clinica.iotools.converters.aibl_to_bids.aibl_to_bids import convert

    convert(
        dataset_directory,
        bids_directory,
        clinical_data_directory,
        overwrite=overwrite,
        clinical_data_only=clinical_data_only,
        n_procs=n_procs,
    )



from pathlib import Path
from typing import Optional

__all__ = [
    "extract_subject_session_longitudinal_ids_from_filename",
    "get_processed_images",
    "init_input_node",
    "read_part_sess_long_ids_from_tsv",
    "run_recon_all_long",
    "save_to_caps",
    "write_tsv_files",
]


def extract_subject_session_longitudinal_ids_from_filename(
    bids_or_caps_files: list[str],
) -> tuple[list[str], list[str], list[str]]:
    """Extract participant/session/longitudinal IDs from filename.

    Example:
        ['sub-CLNC01', 'sub-CLNC01'], ['ses-M000', 'ses-M018'], ['long-M000M018', 'long-M000M018'])

    TODO: Find a way to merge with utils/filemanip.py::extract_subjects_sessions_from_filename into one util
    """
    import re

    id_bids_or_caps_files = [
        re.search(
            r"(sub-[a-zA-Z0-9]+)_(ses-[a-zA-Z0-9]+)_(long-[a-zA-Z0-9]+)", file
        ).group()
        for file in bids_or_caps_files
    ]
    split = [image_id.split("_") for image_id in id_bids_or_caps_files]
    part_ids = [p_id[0] for p_id in split]
    sess_ids = [s_id[1] for s_id in split]
    long_ids = [l_id[2] for l_id in split]
    return part_ids, sess_ids, long_ids


def read_part_sess_long_ids_from_tsv(
    tsv_file: Path,
) -> tuple[list[str], list[str], list[str]]:
    """Extract participant, session and longitudinal from TSV file.

    TODO: Find a way to merge with utils/filemanip.py::read_participant_tsv into one util
    """
    import pandas as pd

    from clinica.utils.exceptions import ClinicaException

    if not tsv_file.is_file():
        raise ClinicaException(
            "The TSV file you gave is not a file.\n"
            "Error explanations:\n"
            f" - Clinica expected the following path to be a file: {tsv_file}\n"
            " - If you gave relative path, did you run Clinica on the good folder?"
        )
    df = pd.read_csv(tsv_file, sep="\t")
    if not {"participant_id", "session_id", "long_id"}.issubset(df.columns):
        raise ClinicaException(
            f"The TSV file ({tsv_file}) does not contain all required columns."
        )

    return (
        [sub.strip(" ") for sub in list(df.participant_id)],
        [ses.strip(" ") for ses in list(df.session_id)],
        [lng.strip(" ") for lng in list(df.long_id)],
    )


def init_input_node(
    caps_dir: Path, participant_id: str, session_id: str, long_id: str, output_dir: Path
) -> Path:
    """Initialize the pipeline."""
    import platform
    from tempfile import mkdtemp

    from clinica.utils.longitudinal import read_sessions
    from clinica.utils.stream import cprint
    from clinica.utils.ux import print_begin_image

    image_id = "_".join((participant_id, session_id, long_id))

    # Create SUBJECTS_DIR for recon-all (otherwise, the command won't run)
    if platform.system().lower().startswith("darwin"):
        # Special case: On macOS, 'recon-all -long' can fail if the $SUBJECTS_DIR is too long
        # To circumvent this issue, we create a sym link in $(TMP) so that $SUBJECTS_DIR is a short path
        subjects_dir = Path(mkdtemp())
        cprint(
            msg=(
                f"Needs to create a $SUBJECTS_DIR folder "
                f"in {subjects_dir} for {image_id.replace('_', ' | ')} (macOS case)."
            ),
            lvl="warning",
        )
    else:
        subjects_dir = output_dir / image_id
    subjects_dir.mkdir(parents=True, exist_ok=True)
    # Create symbolic link containing cross-sectional segmentation(s) in SUBJECTS_DIR so that recon-all can run
    for s_id in read_sessions(caps_dir, participant_id, long_id):
        cross_sectional_path = (
            caps_dir
            / "subjects"
            / participant_id
            / s_id
            / "t1"
            / "freesurfer_cross_sectional"
            / f"{participant_id}_{s_id}"
        )
        (subjects_dir / f"{participant_id}_{s_id}").symlink_to(cross_sectional_path)
    # Create symbolic links containing unbiased template in SUBJECTS_DIR so that recon-all can run
    template_path = (
        caps_dir
        / "subjects"
        / participant_id
        / long_id
        / "freesurfer_unbiased_template"
        / f"{participant_id}_{long_id}"
    )
    (subjects_dir / f"{participant_id}_{long_id}").symlink_to(template_path)

    print_begin_image(image_id)

    return subjects_dir


def run_recon_all_long(
    subjects_dir: str, participant_id: str, session_id: str, long_id: str, directive
) -> str:
    """Run recon-all to create a longitudinal correction of a time point.

    Notes
    -----
    Longitudinal correction with FreeSurfer expects arguments to follow this syntax:
    recon-all -long <tpN_id> <template_id> -all; e.g.: recon-all -long sub-CLNC01_ses-M000 sub-CLNC01_long-M000M018 -all

    Currently, Nipype does not provide interface for "recon-all -long" case. As a result, ReconAll interface should be
    modified to handle this situation. In the meantime, the arguments of this function follows ReconAll.inputs names
    namely:
        - "-long <tpN_id> <template_id>" is likely to be fed to ReconAll.inputs.args
        - "-all" will be fed to ReconAll.inputs.directive

    Folder containing the longitudinal correction has the following convention:
    <tpN_id>.long.<template_id>; e.g.: sub-CLNC01_ses-M000.long.sub-CLNC01_long-M000M018
    which is automatically generated by FreeSurfer.

    This folder name is likely to be retrieved in ReconAll.outputs.subject_id.

    See official documentation (https://surfer.nmr.mgh.harvard.edu/fswiki/LongitudinalProcessing) for details.
    """
    import subprocess

    flags = f" -long {participant_id}_{session_id} {participant_id}_{long_id} "
    recon_all_long_command = f"recon-all {flags} -sd {subjects_dir} {directive}"
    result = subprocess.run(
        recon_all_long_command,
        shell=True,
        stdout=subprocess.DEVNULL,
        stderr=subprocess.DEVNULL,
    )
    if result.returncode != 0:
        raise ValueError("recon-all -long failed, returned non-zero code")

    return f"{participant_id}_{session_id}.long.{participant_id}_{long_id}"


def write_tsv_files(subjects_dir: Path, subject_id: str) -> str:
    """Generate statistics TSV files in `subjects_dir`/regional_measures folder for `subject_id`.

    Notes
    -----
    We do not need to check the line "finished without error" in scripts/recon-all.log.
    If an error occurs, it will be detected by Nipype and the next nodes (i.e.
    write_tsv_files will not be called).
    """
    from clinica.pipelines.anatomical.freesurfer.utils import (
        extract_image_id_from_freesurfer_id,
        generate_regional_measures,
    )
    from clinica.utils.stream import cprint

    if (subjects_dir / subject_id / "mri" / "aparc+aseg.mgz").is_file():
        generate_regional_measures(subjects_dir, subject_id)
    else:
        image_id = " | ".join(extract_image_id_from_freesurfer_id(subject_id))
        cprint(
            msg=(
                f"{image_id} does not contain mri/aseg+aparc.mgz file. "
                "Creation of regional_measures/ folder will be skipped."
            ),
            lvl="warning",
        )
    return subject_id


def save_to_caps(
    source_dir: Path, subject_id: str, caps_dir: Path, overwrite_caps: bool = False
) -> str:
    """Save `source_dir`/`subject_id`/ to CAPS folder.

    This function copies FreeSurfer segmentation and regional_measures folder of `source_dir`/`image_id`/ to
    `caps_dir`/subjects/<participant_id>/<session_id>/<long_id>/t1_freesurfer_longitudinal/
    where `image_id` = <participant_id>_<session_id>_<long_id>.

    The `source_dir`/`image_id`/ folder should contain the following elements:
        - fsaverage, lh.EC_average and rh.EC_average symbolic links automatically generated by recon-all
        - symbolic links to cross-sectional segmentation(s) and unbiased template needed for recon-all
        - <participant_id>_<session_id>.long.<participant_id>_<long_id>/ folder containing the FreeSurfer segmentation
        - regional_measures/ folder containing TSV files

    Notes
    -----
    We do not need to check the line "finished without error" in scripts/recon-all.log.
    If an error occurs, it will be detected by Nipype and the next nodes (i.e. save_to_caps will not be called).
    """
    import shutil

    from clinica.pipelines.anatomical.freesurfer.utils import (
        extract_image_id_from_freesurfer_id,
    )
    from clinica.utils.stream import cprint
    from clinica.utils.ux import print_end_image

    image_id = extract_image_id_from_freesurfer_id(subject_id)
    str_image_id = "_".join(image_id)
    destination_dir = (
        caps_dir.expanduser()
        / "subjects"
        / image_id.participant_id
        / image_id.session_id
        / "t1"
        / image_id.long_id
        / "freesurfer_longitudinal"
    )
    source_file = (
        source_dir.expanduser() / str_image_id / subject_id / "mri" / "aparc+aseg.mgz"
    )
    destination_file = destination_dir / subject_id / "mri" / "aparc+aseg.mgz"
    if source_file.is_file():
        if destination_file.is_file():
            if overwrite_caps:
                shutil.rmtree(destination_dir)
                shutil.copytree(
                    src=source_dir / subject_id / subject_id,
                    dst=destination_dir / subject_id,
                    symlinks=True,
                )
                shutil.copytree(
                    src=source_dir / subject_id / "regional_measures",
                    dst=destination_dir / "regional_measures",
                    symlinks=True,
                )
        else:
            shutil.copytree(
                src=source_dir / str_image_id / subject_id,
                dst=destination_dir / subject_id,
                symlinks=True,
            )
            shutil.copytree(
                src=source_dir / str_image_id / "regional_measures",
                dst=destination_dir / "regional_measures",
                symlinks=True,
            )
        print_end_image(str_image_id)
    else:
        cprint(
            msg=(
                f"{subject_id.replace('_', ' | ')}  does not "
                "contain mri/aseg+aparc.mgz file. Copy will be skipped."
            ),
            lvl="warning",
        )

    return str_image_id


def get_processed_images(
    caps_directory: Path,
    part_ids: list,
    sess_ids: list,
    long_ids: list,
    atlas: Optional[str] = None,
) -> list[str]:
    """
    Extract image IDs (e.g. ["sub-CLNC01_ses-M000_long-M000M018", "sub-CLNC01_ses-M018_long-M000M018"]) of outputs
    already processed by T1FreeSurferLongitudinalCorrection pipeline.
    """
    image_ids = []
    if caps_directory.is_dir():
        for participant_id, session_id, long_id in zip(part_ids, sess_ids, long_ids):
            output_file_pre = (
                caps_directory.expanduser()
                / "subjects"
                / participant_id
                / session_id
                / "t1"
                / long_id
                / "freesurfer_longitudinal"
                / f"{participant_id}_{session_id}.long.{participant_id}_{long_id}"
            )
            if atlas:
                output_file = output_file_pre / "stat" / "rh." / atlas / ".stats"
            else:
                output_file = output_file_pre / "mri" / "aparc+aseg.mgz"
            if output_file.is_file():
                image_ids.append(f"{participant_id}_{session_id}_{long_id}")
    return image_ids



"""This module contains Nipype tasks used in several pipelines."""


def crop_nifti_task(input_image: str, output_path: str) -> str:
    from pathlib import Path

    from clinica.utils.image import crop_nifti

    return str(crop_nifti(Path(input_image), Path(output_path)))


def get_filename_no_ext_task(filename: str) -> str:
    from pathlib import Path

    from clinica.utils.filemanip import get_filename_no_ext

    return get_filename_no_ext(Path(filename))



import json
from pathlib import Path
from typing import IO

from attrs import define, fields
from cattr.gen import make_dict_structure_fn, make_dict_unstructure_fn, override
from cattr.preconf.json import make_converter
from packaging.version import InvalidVersion, Version

from clinica.utils.bids import BIDS_VERSION
from clinica.utils.exceptions import ClinicaBIDSError
from clinica.utils.inputs import DatasetType
from clinica.utils.stream import log_and_raise

__all__ = [
    "BIDSDatasetDescription",
    "get_bids_version",
]


@define
class BIDSDatasetDescription:
    """Model representing a BIDS dataset description.

    See https://bids-specification.readthedocs.io/en/stable/03-modality-agnostic-files.html
    """

    name: str
    bids_version: Version = BIDS_VERSION
    dataset_type: DatasetType = DatasetType.RAW

    def write(self, to: IO[str]):
        import json

        json.dump(converter.unstructure(self), to, indent=4)

    @classmethod
    def from_file(cls, json_file: Path):
        with open(json_file, "r") as fp:
            content = json.load(fp)
        return converter.structure(content, BIDSDatasetDescription)


def _rename(name: str) -> str:
    """Rename attributes following the specification for the JSON file.

    Basically pascal case with known acronyms such as BIDS fully capitalized.
    """
    return "".join(
        word.upper() if word == "bids" else word.capitalize()
        for word in name.split("_")
    )


# Register a JSON converter for the BIDS dataset description model.
converter = make_converter()
converter.register_unstructure_hook(Version, lambda dt: str(dt))
converter.register_structure_hook(Version, lambda ts, _: Version(ts))
bids_dataset_description_field_renaming = {
    a.name: override(rename=_rename(a.name)) for a in fields(BIDSDatasetDescription)
}
converter.register_unstructure_hook(
    BIDSDatasetDescription,
    make_dict_unstructure_fn(
        BIDSDatasetDescription,
        converter,
        **bids_dataset_description_field_renaming,
    ),
)
bids_dataset_field_renaming_structure_hook = make_dict_structure_fn(
    BIDSDatasetDescription,
    converter,
    **bids_dataset_description_field_renaming,
)
converter.register_structure_hook(
    BIDSDatasetDescription,
    bids_dataset_field_renaming_structure_hook,
)


def get_bids_version(dataset_folder: Path) -> Version:
    """Returns the BIDS version number of a BIDS or CAPS dataset."""
    try:
        with open(dataset_folder / "dataset_description.json", "r") as fp:
            bids_metadata = json.load(fp)
        return Version(bids_metadata["BIDSVersion"])
    except InvalidVersion as e:
        log_and_raise(
            (
                f"File {dataset_folder / 'dataset_description.json'} has a "
                f"BIDS version number not properly formatted:\n{e}"
            ),
            ClinicaBIDSError,
        )
    except FileNotFoundError:
        log_and_raise(
            (
                f"File {dataset_folder / 'dataset_description.json'} is missing "
                "while it is mandatory for a BIDS/CAPS dataset."
            ),
            ClinicaBIDSError,
        )
    except KeyError:
        log_and_raise(
            (
                f"File {dataset_folder / 'dataset_description.json'} is missing a "
                "'BIDSVersion' key while it is mandatory."
            ),
            ClinicaBIDSError,
        )
    except json.JSONDecodeError as e:
        log_and_raise(
            f"File {dataset_folder / 'dataset_description.json'} is not formatted correctly:\n{e}.",
            ClinicaBIDSError,
        )



"""Convert IXI dataset (https://brain-development.org/ixi-dataset/) to BIDS."""

from pathlib import Path
from typing import Optional

import nibabel as nb
import numpy as np

from clinica.iotools.bids_utils import write_modality_agnostic_files
from clinica.iotools.converters.ixi_to_bids.ixi_to_bids_utils import (
    check_modalities,
    define_participants,
    read_clinical_data,
    write_participants,
    write_scans,
    write_sessions,
    write_subject_data,
)
from clinica.utils.filemanip import UserProvidedPath

__all__ = ["convert"]


def convert(
    path_to_dataset: UserProvidedPath,
    bids_dir: UserProvidedPath,
    path_to_clinical: UserProvidedPath,
    subjects: Optional[UserProvidedPath] = None,
    n_procs: Optional[int] = 1,
    **kwargs,
):
    from clinica.iotools.bids_utils import StudyName
    from clinica.iotools.converters.factory import get_converter_name
    from clinica.utils.stream import cprint

    from ..utils import validate_input_path

    path_to_dataset = validate_input_path(path_to_dataset)
    bids_dir = validate_input_path(bids_dir, check_exist=False)
    path_to_clinical = validate_input_path(path_to_clinical)
    if subjects:
        subjects = validate_input_path(subjects)

    if n_procs != 1:
        cprint(
            f"{get_converter_name(StudyName.IXI)} converter does not support multiprocessing yet. n_procs set to 1.",
            lvl="warning",
        )

    clinical_data = read_clinical_data(path_to_clinical)
    participants = define_participants(path_to_dataset, subjects)
    check_modalities(data_directory=path_to_dataset, participants=participants)

    write_participants(
        bids_dir=bids_dir, clinical_data=clinical_data, participants=participants
    )

    for participant in participants:
        cprint(f"Converting IXI subject {participant} to BIDS", lvl="debug")
        write_subject_data(
            bids_dir=bids_dir, participant=participant, path_to_dataset=path_to_dataset
        )
        write_sessions(
            bids_dir=bids_dir, participant=participant, clinical_data=clinical_data
        )
        write_scans(bids_dir=bids_dir, participant=participant)

    readme_data = {
        "link": "https://brain-development.org/ixi-dataset/",
        "desc": (
            "IXI is the nickname for the Information eXtraction from Images project, "
            "which issued a dataset of nearly 600 images from healthy subjects. The MR"
            "acquisition protocol includes T1,T2, PD weighted, MRA and diffusion-weighted"
            "images. Three hospitals in London were involved in data collection."
        ),
    }

    write_modality_agnostic_files(
        study_name=StudyName.IXI, readme_data=readme_data, bids_dir=bids_dir
    )

    cprint("Conversion to BIDS finished.", lvl="info")



from os import PathLike
from pathlib import Path
from typing import Optional

from pydra.mark import annotate, task


@task
@annotate(
    {
        "return": {
            "bids_file": Path,
            "participant_id": str,
            "session_id": Optional[str],
            "datatype": str,
            "suffix": str,
            "extension": str,
            "entities": dict,
        }
    }
)
def parse_bids_file(bids_file: PathLike):
    """Parse the components of a BIDS file.

    Examples
    --------
    >>> task = parse_bids_file(bids_file=Path("dataset/sub-P01/ses-M00/anat/sub-P01_ses-M00_desc-test_T1w.nii.gz"))
    >>> result = task()
    >>> result.output.participant_id
    'sub-P01'
    >>> result.output.session_id
    'ses-M00'
    >>> result.output.datatype
    'anat'
    >>> result.output.suffix
    'T1w'
    >>> result.output.extension
    'nii.gz'
    >>> result.output.entities
    {'desc': 'test'}
    """
    bids_file = Path(bids_file)

    datatype = bids_file.parent.name
    file_name = bids_file.name
    rest, extension = file_name.split(sep=".", maxsplit=1)
    rest, suffix = rest.rsplit(sep="_", maxsplit=1)
    entities = {
        prefix: value for prefix, value in (part.split("-") for part in rest.split("_"))
    }

    participant_id = f"sub-{entities.pop('sub')}"
    try:
        session_id = f"ses-{entities.pop('ses')}"
    except KeyError:
        session_id = None

    return bids_file, participant_id, session_id, datatype, suffix, extension, entities



"""Convert the GENFI dataset into BIDS."""

from pathlib import Path
from typing import Optional

from clinica.utils.filemanip import UserProvidedPath

__all__ = ["convert"]


def convert(
    path_to_dataset: UserProvidedPath,
    bids_dir: UserProvidedPath,
    path_to_clinical: Optional[UserProvidedPath] = None,
    gif: Optional[bool] = False,
    path_to_clinical_tsv: Optional[UserProvidedPath] = None,
    subjects: Optional[UserProvidedPath] = None,
    n_procs: Optional[int] = 1,
    **kwargs,
) -> None:
    """Convert the entire dataset to BIDS.

    Scans available files in the path_to_dataset,
    identifies the patients that have images described by the JSON file,
    converts the image with the highest quality for each category.

    Parameters
    ----------
    path_to_dataset: Path
        The path to the raw images.

    bids_dir: Path
        The path to directory where the bids will be written.

    path_to_clinical: Path, optional
        The path to the clinical data associated with the dataset.
        If None, the clinical data won't be converted.

    gif: bool, optional
        If True, indicates the user wants to have the values of the gif parcellation

    path_to_clinical_tsv: Path, optional
        The path to a TSV file containing the additional data the user wants to have in the BIDS output.
        If None, no additional data will be added.

    subjects : str or Path, optional
        The path to a file defining a subset of subjects to be converted.
        By default, all subjects available will be converted.

    n_procs : int, optional
        The requested number of processes.
        If specified, it should be between 1 and the number of available CPUs.
        Default=1.
    """
    from clinica.iotools.bids_utils import StudyName, write_modality_agnostic_files
    from clinica.iotools.converters.factory import get_converter_name
    from clinica.utils.check_dependency import ThirdPartySoftware, check_software
    from clinica.utils.stream import cprint

    from ..utils import validate_input_path
    from .genfi_to_bids_utils import (
        merge_imaging_and_clinical_data,
        parse_clinical_data,
        parse_imaging_data,
        prepare_dataset_to_bids_format,
        write_bids,
    )

    path_to_dataset = validate_input_path(path_to_dataset)
    bids_dir = validate_input_path(bids_dir, check_exist=False)
    path_to_clinical = validate_input_path(path_to_clinical)
    if path_to_clinical_tsv:
        path_to_clinical_tsv = validate_input_path(path_to_clinical_tsv)
    check_software(ThirdPartySoftware.DCM2NIIX)
    _check_clinical_path_inputs(path_to_clinical_tsv, path_to_clinical)
    if subjects:
        cprint(
            (
                f"Subject filtering is not yet implemented in {get_converter_name(StudyName.GENFI)} converter. "
                "All subjects available will be converted."
            ),
            lvl="warning",
        )
    if n_procs != 1:
        cprint(
            f"{get_converter_name(StudyName.GENFI)} converter does not support multiprocessing yet. n_procs set to 1.",
            lvl="warning",
        )
    imaging_data = parse_imaging_data(path_to_dataset)
    if path_to_clinical:
        clinical_data = parse_clinical_data(path_to_clinical)
        imaging_data = merge_imaging_and_clinical_data(imaging_data, clinical_data)
    results = prepare_dataset_to_bids_format(imaging_data, gif, path_to_clinical_tsv)
    write_bids(
        to=bids_dir,
        participants=results["participants"],
        sessions=results["sessions"],
        scans=results["scans"],
    )
    write_modality_agnostic_files(
        study_name=StudyName.GENFI,
        readme_data={
            "link": _get_link(),
            "desc": _get_description(),
        },
        bids_dir=bids_dir,
    )
    cprint("Conversion to BIDS succeeded.", lvl="info")


def _check_clinical_path_inputs(path_to_clinical_tsv: Path, path_to_clinical: Path):
    """Check that if a clinical tsv is given, a path to the clinical data is given as well."""
    from clinica.iotools.bids_utils import StudyName
    from clinica.iotools.converters.factory import get_converter_name
    from clinica.utils.stream import cprint

    if path_to_clinical_tsv and not path_to_clinical:
        msg = (
            f"The {get_converter_name(StudyName.GENFI)} converter is unable to convert the clinical data because "
            "the path to these data was not provided while a TSV file with additional "
            f"data was given ({path_to_clinical_tsv}). You can either use the appropriate "
            "option from the clinica command line interface to provide the missing path, "
            "or chose to not convert clinical data at all."
        )
        cprint(msg, lvl="error")
        raise ValueError(msg)


def _get_link() -> str:
    return "https://www.genfi.org"


def _get_description() -> str:
    return (
        "The Genetic Frontotemporal dementia Initiative (GENFI) is a group of research "
        "centres across Europe and Canada with expertise in familial FTD, and is "
        "co-ordinated by Professor Jonathan Rohrer at University College London. "
        "GENFI is the largest genetic FTD consortium to date and currently consists "
        "of sites across the UK, Netherlands, Belgium, France, Spain, Portugal, Italy, "
        "Germany, Sweden, Denmark, Finland and Canada. The aim of the study is to "
        "understand more about genetic FTD, particularly in those who have mutations "
        "in the progranulin (GRN), microtubule-associated protein tau (MAPT) and "
        "chromosome 9 open reading frame 72 (C9orf72) genes. GENFI investigates both "
        "people who have developed symptoms and also people who have a risk of developing "
        "symptoms in the future because they carry an abnormal genetic mutation. "
        "By studying these individuals who are destined to develop the disease later in "
        "life we can understand the development from the very earliest changes. "
        "The key objectives of GENFI are therefore to develop markers which help identify "
        "the disease at its earliest stage as well as markers that allow the progression of "
        "the disease to be tracked. We are now collaborating closely with other similar "
        "studies around the world through the FTD Prevention Initiative. "
        "Through this worldwide initiative we are working with pharmaceutical companies to "
        "help design clinical trials for genetic FTD."
    )



from typing import List, Optional, Union

import click

from clinica import option
from clinica.pipelines import cli_param
from clinica.pipelines.engine import clinica_pipeline
from clinica.utils.pet import ReconstructionMethod, SUVRReferenceRegion, Tracer

pipeline_name = "pet-volume"


@clinica_pipeline
@click.command(name=pipeline_name)
@cli_param.argument.bids_directory
@cli_param.argument.caps_directory
@cli_param.argument.group_label
@cli_param.argument.acq_label
@cli_param.argument.suvr_reference_region
@cli_param.option_group.pipeline_specific_options
@cli_param.option.pvc_psf_tsv
@cli_param.option_group.common_pipelines_options
@cli_param.option.reconstruction_method
@cli_param.option.subjects_sessions_tsv
@cli_param.option.working_directory
@cli_param.option.yes
@option.global_option_group
@option.n_procs
@cli_param.option_group.advanced_pipeline_options
@cli_param.option_group.option(
    "-mask",
    "--mask_tissues",
    multiple=True,
    type=click.IntRange(1, 7),
    default=[1, 2, 3],
    help=(
        "Tissue classes (1: gray matter (GM), 2: white matter (WM), 3: cerebrospinal "
        "fluid (CSF), 4: bone, 5: soft-tissue, 6: background) to use "
        "for masking the PET image (default: GM, WM and CSF i.e. --mask_tissues 1 2 3)."
    ),
)
@cli_param.option_group.option(
    "-threshold",
    "--mask_threshold",
    default=0.3,
    show_default=True,
    help="Value used as threshold to binarize the tissue maps.",
)
@cli_param.option_group.option(
    "-pvc_mask",
    "--pvc_mask_tissues",
    multiple=True,
    type=click.IntRange(1, 7),
    default=[1, 2, 3],
    help=(
        "Tissue classes (1: gray matter (GM), 2: white matter (WM), 3: cerebrospinal "
        "fluid (CSF), 4: bone, 5: soft-tissue, 6: background) to use "
        "as mask for PVC (default: GM, WM and CSF i.e. --pvc_mask_tissues 1 2 3)."
    ),
)
@cli_param.option.smooth
def cli(
    bids_directory: str,
    caps_directory: str,
    group_label: str,
    acq_label: Union[str, Tracer],
    suvr_reference_region: Optional[Union[str, SUVRReferenceRegion]] = None,
    reconstruction_method: Optional[Union[str, ReconstructionMethod]] = None,
    pvc_psf_tsv: Optional[str] = None,
    mask_tissues: List[int] = (1, 2, 3),
    mask_threshold: float = 0.3,
    pvc_mask_tissues: List[int] = (1, 2, 3),
    smooth: List[int] = (8, 8),
    subjects_sessions_tsv: Optional[str] = None,
    working_directory: Optional[str] = None,
    n_procs: Optional[int] = None,
    yes: bool = False,
) -> None:
    """SPM-based pre-processing of PET images.

       GROUP_LABEL is an user-defined identifier to target a specific group of subjects.
    For this pipeline, it is associated to the DARTEL template that you had created when running the t1-volume pipeline.

       ACQ_LABEL corresponds to the label given to the PET acquisition, specifying the tracer used.
    Frequently used values are '18FFDG' or '18FAV45'.

       The reference region must be specified to perform intensity normalization.
    Accepted values include: 'pons', 'cerebellumPons', 'pons2', 'cerebellumPons2'.

    Prerequisite: You need to have performed the t1-volume pipeline on your T1-weighted MR images.

    See https://aramislab.paris.inria.fr/clinica/docs/public/latest/Pipelines/PET_Volume/
    """
    from networkx import Graph

    from clinica.utils.ux import print_end_pipeline

    from .pipeline import PETVolume

    parameters = {
        "group_label": group_label,
        "acq_label": acq_label,
        "suvr_reference_region": suvr_reference_region,
        "reconstruction_method": reconstruction_method,
        "pvc_psf_tsv": pvc_psf_tsv,
        "mask_tissues": mask_tissues,
        "mask_threshold": mask_threshold,
        "pvc_mask_tissues": pvc_mask_tissues,
        "smooth": smooth,
        "skip_question": yes,
    }

    pipeline = PETVolume(
        bids_directory=bids_directory,
        caps_directory=caps_directory,
        tsv_file=subjects_sessions_tsv,
        base_dir=working_directory,
        parameters=parameters,
        name=pipeline_name,
    )

    exec_pipeline = (
        pipeline.run(plugin="MultiProc", plugin_args={"n_procs": n_procs})
        if n_procs
        else pipeline.run()
    )

    if isinstance(exec_pipeline, Graph):
        print_end_pipeline(
            pipeline_name, pipeline.base_dir, pipeline.base_dir_was_specified
        )


if __name__ == "__main__":
    cli()



"""This module contains utilities used by the DWIPreprocessingUsingPhaseDiff pipeline."""

import math
from pathlib import Path
from typing import Optional, Tuple

import nibabel as nib
import numpy as np

__all__ = [
    "rename_into_caps",
    "get_grad_fsl",
    "init_input_node",
    "print_end_pipeline",
    "convert_phase_difference_to_hertz",
    "demean_image",
    "convert_phase_difference_to_rads",
]


def rename_into_caps(
    dwi_filename: Path,
    dwi_preproc_filename: Path,
    b_values_preproc_filename: Path,
    b_vectors_preproc_filename: Path,
    b0_brain_mask_filename: Path,
    calibrated_magnitude_image_filename: Path,
    calibrated_field_map_image_filename: Path,
    calibrated_smoothed_field_map_image_filename: Path,
) -> Tuple[str, ...]:
    """Rename the outputs of the pipelines into CAPS.

    Parameters
    ----------
    dwi_filename : Path
        The path to input BIDS DWI to extract the <source_file>.

    dwi_preproc_filename : Path
        The path to the preprocessed DWI file.

    b_values_preproc_filename : Path
        The path to the preprocessed b-values file.

    b_vectors_preproc_filename : Path
        The path to the preprocessed b-vectors file.

    b0_brain_mask_filename : Path
        The path to the B0 mask file.

    calibrated_magnitude_image_filename : Path
        The path to the magnitude image file on b0 space.

    calibrated_field_map_image_filename : Path
        The path to the calibrated fmap file on b0 space.

    calibrated_smoothed_field_map_image_filename : Path
        The path to the smoothed (calibrated) fmap file on b0 space.

    Returns
    -------
    Tuple[str, str, str, str, str, str, str] :
        The different outputs in CAPS format.
    """
    from clinica.pipelines.dwi.utils import rename_files

    return rename_files(
        dwi_filename,
        {
            dwi_preproc_filename: "_space-b0_desc-preproc_dwi.nii.gz",
            b_values_preproc_filename: "_space-b0_desc-preproc_dwi.bval",
            b_vectors_preproc_filename: "_space-b0_desc-preproc_dwi.bvec",
            b0_brain_mask_filename: "_space-b0_brainmask.nii.gz",
            calibrated_magnitude_image_filename: "_space-b0_magnitude1.nii.gz",
            calibrated_field_map_image_filename: "_space-b0_fmap.nii.gz",
            calibrated_smoothed_field_map_image_filename: "_space-b0_fwhm-4_fmap.nii.gz",
        },
    )


def get_grad_fsl(b_vectors_filename: str, b_values_filename: str) -> tuple:
    return b_vectors_filename, b_values_filename


def init_input_node(
    dwi_filename: str,
    b_vectors_filename: str,
    b_values_filename: str,
    dwi_json_filename: str,
    fmap_magnitude_filename: str,
    fmap_phasediff_filename: str,
    fmap_phasediff_json_filename: str,
) -> tuple:
    """Initialize pipeline (read JSON, check files and print begin message).

    Parameters
    ----------
    dwi_filename : str
        The path to the DWI image.

    b_vectors_filename : str
        The path to the b-vectors file.

    b_values_filename : str
        The path to the b-values file.

    dwi_json_filename : str
        Path to the JSON metadata file for the DWI image.

    fmap_magnitude_filename : str
        The path to the (1st) magnitude image in BIDS format.

    fmap_phasediff_filename : str
        The path of the phase difference image in BIDS format.

    fmap_phasediff_json_filename : str
        The path of the phase difference JSON file in BIDS format
        and containing EchoTime1 & EchoTime2 metadata (see BIDS specifications).

    Returns
    -------
    image_id : str
        The subject ID extracted from the dwi image path.

    dwi_filename : str
        The path to the DWI image.

    b_vectors_filename : str
        The path to the b-vectors file.

    b_values_filename : str
        The path to the b-values file.

    total_readout_time : str
        The total readout time extracted from the dwi JSON file.

    phase_encoding_direction : str
        The phase encoding direction for the dwi image, extracted
        from the dwi JSON file.

    fmap_magnitude_filename : str
        The path to the (1st) magnitude image in BIDS format.

    fmap_phasediff_filename : str
        The path of the phase difference image in BIDS format.

    delta_echo_time : str
        The path to the JSON metadata file for the field map image.
    """
    import datetime

    import nibabel as nib

    from clinica.pipelines.dwi.preprocessing.utils import (
        check_dwi_volume,
        get_readout_time_and_phase_encoding_direction,
    )
    from clinica.pipelines.dwi.utils import DWIDataset
    from clinica.utils.filemanip import extract_metadata_from_json, get_subject_id
    from clinica.utils.stream import cprint
    from clinica.utils.ux import print_begin_image

    image_id = get_subject_id(dwi_filename)
    try:
        check_dwi_volume(
            DWIDataset(
                dwi=dwi_filename,
                b_values=b_values_filename,
                b_vectors=b_vectors_filename,
            )
        )
    except ValueError as e:
        now = datetime.datetime.now().strftime("%H:%M:%S")
        error_msg = f"[{now}] Error: Number of DWIs, b-vals and b-vecs mismatch for {image_id.replace('_', ' | ')}"
        cprint(error_msg, lvl="error")
        raise ValueError(e)

    # Check that PhaseDiff and magnitude1 have the same header
    # Otherwise, FSL in FugueExtrapolationFromMask will crash
    img_phasediff = nib.load(fmap_phasediff_filename)
    img_magnitude = nib.load(fmap_magnitude_filename)
    if img_phasediff.shape != img_magnitude.shape:
        now = datetime.datetime.now().strftime("%H:%M:%S")
        error_msg = (
            f"[{now}] Error: Headers of PhaseDiff and Magnitude1 are not the same "
            f"for {image_id.replace('_', ' | ')} ({img_phasediff.shape} vs {img_magnitude.shape})"
        )
        cprint(error_msg, lvl="error")
        raise NotImplementedError(error_msg)
    (
        total_readout_time,
        phase_encoding_direction,
    ) = get_readout_time_and_phase_encoding_direction(dwi_json_filename)

    # Read metadata from PhaseDiff JSON file:
    [echo_time1, echo_time2] = extract_metadata_from_json(
        fmap_phasediff_json_filename, ["EchoTime1", "EchoTime2"]
    )
    delta_echo_time = abs(echo_time2 - echo_time1)

    print_begin_image(
        image_id,
        ["TotalReadoutTime", "PhaseEncodingDirection", "DeltaEchoTime"],
        [str(total_readout_time), phase_encoding_direction, str(delta_echo_time)],
    )

    return (
        image_id,
        dwi_filename,
        b_vectors_filename,
        b_values_filename,
        total_readout_time,
        phase_encoding_direction,
        fmap_magnitude_filename,
        fmap_phasediff_filename,
        delta_echo_time,
    )


def print_end_pipeline(image_id, final_file):
    """Display end message for `image_id` when `final_file` is connected."""
    from clinica.utils.ux import print_end_image

    print_end_image(image_id)


def convert_phase_difference_to_hertz(
    phase_diff_filename: Path,
    delta_echo_time: float,
    working_dir: Optional[Path] = None,
) -> Path:
    """Convert input phase difference map to Hz.

    Parameters
    ----------
    phase_diff_filename : Path
        The path to the phase difference map image.

    delta_echo_time : float
        The DeltaEchoTime from BIDS specifications.

    working_dir : Path, optional
        The path to the working directory. If None, the output image
        will be written in current folder. The file name
        will have the same base name as in_file, but with
        the "_radsec.nii.gz" suffix.

    Returns
    -------
    out_file : Path
        The path to output file.
    """
    im = nib.load(phase_diff_filename)
    working_dir = working_dir or Path.cwd()
    out_file = working_dir / _get_output_file(phase_diff_filename, "radsec")
    data = im.get_fdata().astype(np.float32) * (
        1.0 / (float(delta_echo_time) * 2 * math.pi)
    )
    nib.Nifti1Image(data, im.affine, im.header).to_filename(out_file)

    return out_file


def _get_output_file(input_file: Path, suffix: str) -> str:
    filename = input_file.stem
    if input_file.suffix == ".gz":
        filename = Path(filename).stem

    return f"{filename}_{suffix}.nii.gz"


def demean_image(
    input_image: Path, mask: Optional[Path] = None, working_dir: Optional[Path] = None
) -> Path:
    """Demean image data inside mask.

    This function was taken from: https://github.com/niflows/nipype1-workflows/

    Parameters
    ----------
    input_image : Path
        The image to demean.

    mask : Path, optional
        If provided, will be used to mask the image.

    working_dir : Path, optional
        The path to the working directory. If None, the output image
        will be written in current folder. The file name
        will have the same base name as in_file, but with
        the "_demean.nii.gz" suffix.

    Returns
    -------
    out_file : Path
        The path to output file.
    """
    image = nib.load(input_image)
    working_dir = working_dir or Path.cwd()
    out_file = working_dir / _get_output_file(input_image, "demean")
    data = image.get_fdata().astype(np.float32)
    if mask:
        mask = nib.load(mask).get_fdata().astype(np.float32)
        mask[mask > 0] = 1.0
        mask[mask < 1] = 0.0
    else:
        mask = np.ones_like(data)
    mean_image = np.median(data[mask == 1].reshape(-1))
    data[mask == 1] = data[mask == 1] - mean_image
    nib.Nifti1Image(data, image.affine, image.header).to_filename(out_file)

    return out_file


def convert_phase_difference_to_rads(
    phase_diff_filename: Path, working_dir: Optional[Path] = None
) -> Path:
    """Converts input phase difference map to rads.

    This function was taken from: https://github.com/niflows/nipype1-workflows/

    Parameters
    ----------
    phase_diff_filename : Path
        The path to the phase difference map image.

    working_dir : Path, optional
        The path to the working directory. If None, the output image
        will be written in current folder. The file name
        will have the same base name as in_file, but with
        the "_rads.nii.gz" suffix.

    Returns
    -------
    out_file : Path
        The path to output file.
    """
    working_dir = working_dir or Path.cwd()
    out_file = working_dir / _get_output_file(phase_diff_filename, "rads")
    in_file = np.atleast_1d(phase_diff_filename).tolist()
    image = nib.load(in_file[0])
    data = image.get_fdata().astype(np.float32)
    header = image.header.copy()

    if len(in_file) == 2:
        data = nib.load(in_file[1]).get_fdata().astype(np.float32) - data
    elif (data.ndim == 4) and (data.shape[-1] == 2):
        data = np.squeeze(data[..., 1] - data[..., 0])
        header.set_data_shape(data.shape[:3])

    data_min = data.min()
    data_max = data.max()
    data = (2.0 * math.pi * (data - data_min) / (data_max - data_min)) - math.pi
    header.set_data_dtype(np.float32)
    header.set_xyzt_units("mm")
    header["datatype"] = 16
    nib.Nifti1Image(data, image.affine, header).to_filename(out_file)

    return out_file



"""Module for converting Tau PET of ADNI."""

from pathlib import Path
from typing import Iterable, Optional

import pandas as pd

__all__ = ["convert_tau_pet"]


def convert_tau_pet(
    source_dir: Path,
    csv_dir: Path,
    destination_dir: Path,
    conversion_dir: Path,
    subjects: Iterable[str],
    mod_to_update: bool = False,
    n_procs: int = 1,
):
    """Convert Tau PET images of ADNI into BIDS format.

    Parameters
    ----------
    source_dir : Path
        The path to the ADNI directory.

    csv_dir : Path
        The path to the clinical data directory.

    destination_dir : Path
        The path to the destination BIDS directory.

    conversion_dir : Path
        The path to the TSV files including the paths to original images.

    subjects : List of str, optional
        List of subjects.

    mod_to_update : bool
        If True, pre-existing images in the BIDS directory
        will be erased and extracted again.

    n_procs : int, optional
        The requested number of processes.
        If specified, it should be between 1 and the number of available CPUs.
        Default=1
    """
    from clinica.iotools.converters.adni_to_bids.adni_utils import (
        ADNIModalityConverter,
        load_clinical_csv,
        paths_to_bids,
    )
    from clinica.utils.stream import cprint

    cprint(
        (
            f"Calculating paths of {ADNIModalityConverter.PET_TAU.value} images. "
            f"Output will be stored in {conversion_dir}."
        ),
        lvl="info",
    )
    images = _compute_tau_pet_paths(source_dir, csv_dir, subjects, conversion_dir)
    cprint(
        f"Paths of {ADNIModalityConverter.PET_TAU.value} images found. Exporting images into BIDS ...",
        lvl="info",
    )
    paths_to_bids(
        images,
        destination_dir,
        ADNIModalityConverter.PET_TAU,
        mod_to_update=mod_to_update,
        n_procs=n_procs,
    )
    cprint(msg=f"{ADNIModalityConverter.PET_TAU.value} conversion done.", lvl="debug")


def _compute_tau_pet_paths(
    source_dir: Path,
    csv_dir: Path,
    subjects: Iterable[str],
    conversion_dir: Path,
) -> pd.DataFrame:
    """Compute the paths to Tau PET images.

    Parameters
    ----------
    source_dir : Path
        The path to the ADNI directory.

    csv_dir : Path
        The path to the clinical data directory.

    subjects : list of str
        The subjects list.

    conversion_dir : Path
        The path to the TSV files including the paths to original images.

    Returns
    -------
    pd.DataFrame :
        A pandas Dataframe containing the path for each Tau PET image.
    """
    from clinica.iotools.converters.adni_to_bids.adni_utils import load_clinical_csv
    from clinica.utils.pet import Tracer

    from ._image_path_utils import find_image_path
    from ._pet_utils import get_images_pet

    pet_tau_df = pd.DataFrame(columns=_get_tau_pet_df_columns())
    pet_tau_dfs_list = []
    tauqc = load_clinical_csv(csv_dir, "TAUQC")
    tauqc3 = load_clinical_csv(csv_dir, "TAUQC3")
    pet_meta_list = load_clinical_csv(csv_dir, "PET_META_LIST")

    for subject in subjects:
        # PET images metadata for subject
        subject_pet_meta = pet_meta_list[pet_meta_list["Subject"] == subject]
        if subject_pet_meta.empty:
            continue
        # QC for TAU PET images for ADNI 2
        tau_qc2_subj = tauqc[(tauqc.SCANQLTY == 1) & (tauqc.RID == int(subject[-4:]))]
        # QC for TAU PET images for ADNI 3
        tau_qc3_subj = tauqc3[
            (tauqc3.SCANQLTY == 1) & (tauqc3.RID == int(subject[-4:]))
        ]
        # Concatenating visits in both QC files
        tau_qc_subj = pd.concat(
            [tau_qc2_subj, tau_qc3_subj], axis=0, ignore_index=True, sort=False
        )
        tau_qc_subj.rename(columns={"SCANDATE": "EXAMDATE"}, inplace=True)
        subj_dfs_list = get_images_pet(
            subject=subject,
            pet_qc_subj=tau_qc_subj,
            subject_pet_meta=subject_pet_meta,
            df_cols=_get_tau_pet_df_columns(),
            modality="TAU-PET",
            sequences_preprocessing_step=["AV1451 Co-registered, Averaged"],
        )
        if subj_dfs_list:
            pet_tau_dfs_list += subj_dfs_list
    if pet_tau_dfs_list:
        pet_tau_df = pd.concat(pet_tau_dfs_list, ignore_index=True)

    # Removing known exceptions from images to convert
    if not pet_tau_df.empty:
        error_ind = pet_tau_df.index[
            pet_tau_df.apply(
                lambda x: ((x.Subject_ID, x.VISCODE) in _get_known_conversion_errors()),
                axis=1,
            )
        ]
        pet_tau_df.drop(error_ind, inplace=True)

    # Checking for images paths in filesystem
    images = find_image_path(pet_tau_df, source_dir, "TAU")
    images.to_csv(
        conversion_dir / f"{Tracer.AV1451.value}_pet_paths.tsv",
        sep="\t",
        index=False,
    )

    return images


def _get_tau_pet_df_columns() -> list[str]:
    return [
        "Phase",
        "Subject_ID",
        "VISCODE",
        "Visit",
        "Sequence",
        "Scan_Date",
        "Study_ID",
        "Series_ID",
        "Image_ID",
        "Original",
    ]


def _get_known_conversion_errors() -> Iterable[tuple[str, str]]:
    return [("098_S_4275", "m84")]



"""
Convert the AIBL dataset (https://www.aibl.csiro.au/) into BIDS.
"""
from pathlib import Path
from typing import Optional

from clinica.utils.filemanip import UserProvidedPath

__all__ = ["convert"]


def convert(
    input_dataset: UserProvidedPath,
    output_dataset: UserProvidedPath,
    input_clinical_data: UserProvidedPath,
    overwrite: bool = False,
    clinical_data_only: bool = False,
    subjects: Optional[UserProvidedPath] = None,
    n_procs: Optional[int] = 1,
    **kwargs,
) -> None:
    """Convert the AIBL dataset (https://www.aibl.csiro.au/) in a BIDS dataset.

    Parameters
    ----------
    input_dataset : Path
        The path to the input AIBL dataset.

    output_dataset : Path
        The path to the BIDS directory in which to write the output.

    input_clinical_data : Path
        The path to the clinical CSV files associated with the input dataset.

    overwrite : bool, optional
        Overwrites previously written nifti and json files.
        Default=False.

    clinical_data_only : bool, optional
        If True, only convert the clinical data without the imaging data.
        If False, everything is converted.
        Default=False.

    subjects : str or Path, optional
        The path to a file defining a subset of subjects to be converted.
        By default, all subjects available will be converted.

    n_procs : int, optional
        The requested number of processes.
        If specified, it should be between 1 and the number of available CPUs.
        Default=1.
    """
    from clinica.iotools.bids_utils import StudyName
    from clinica.iotools.converters.factory import get_converter_name
    from clinica.utils.check_dependency import ThirdPartySoftware, check_software
    from clinica.utils.stream import cprint

    from ..utils import validate_input_path

    input_dataset = validate_input_path(input_dataset)
    output_dataset = validate_input_path(output_dataset, check_exist=False)
    input_clinical_data = validate_input_path(input_clinical_data)
    check_software(ThirdPartySoftware.DCM2NIIX)
    if subjects:
        cprint(
            (
                f"Subject filtering is not yet implemented in {get_converter_name(StudyName.AIBL)} converter. "
                "All subjects available will be converted."
            ),
            lvl="warning",
        )
    output_dataset.mkdir(parents=True, exist_ok=True)
    if not clinical_data_only:
        _convert_images(
            input_dataset,
            input_clinical_data,
            output_dataset,
            overwrite,
            n_procs=n_procs,
        )
    _convert_clinical_data(input_clinical_data, output_dataset)


def _convert_images(
    input_dataset: Path,
    input_clinical_data: Path,
    output_dataset: Path,
    overwrite: bool = False,
    n_procs: Optional[int] = 1,
) -> None:
    """Conversion of the AIBL imaging data in BIDS.

    Parameters
    ----------
    input_dataset : Path
        The path to the input AIBL dataset.

    input_clinical_data : Path
        The path to the clinical CSV files associated with the input dataset.

    output_dataset : Path
        The path to the BIDS directory in which to write the output.

    overwrite : bool, optional
        Overwrites previously written nifti and json files.
        Default=False.

    n_procs : int, optional
        The requested number of processes.
        If specified, it should be between 1 and the number of available CPUs.
        Default=1.
    """
    from clinica.iotools.converters.aibl_to_bids.utils import Modality, paths_to_bids

    created_files = [
        paths_to_bids(
            input_dataset,
            input_clinical_data,
            output_dataset,
            modality,
            overwrite=overwrite,
            n_procs=n_procs,
        )
        for modality in Modality
    ]
    _warn_about_missing_files(created_files)


def _warn_about_missing_files(files: list[list[Optional[Path]]]):
    from clinica.utils.stream import cprint

    missing_files = []
    for files_for_given_modality in files:
        for file in files_for_given_modality:
            if file is not None and not file.exists():
                missing_files.append(file)
    if missing_files:
        msg = "The following file were not converted:\n" + "\n".join(
            (str(f) for f in missing_files)
        )
        cprint(msg=msg, lvl="warning")


def _convert_clinical_data(input_clinical_data: Path, output_dataset: Path) -> None:
    """Conversion of the AIBL clinical data in BIDS.

    Parameters
    ----------
    input_clinical_data : Path
        The path to the clinical CSV files associated with the input dataset.

    output_dataset : Path
        The path to the BIDS directory in which to write the output.
    """
    from clinica.iotools.bids_utils import StudyName, write_modality_agnostic_files
    from clinica.iotools.converters.aibl_to_bids.utils import (
        create_participants_tsv_file,
        create_scans_tsv_file,
        create_sessions_tsv_file,
    )
    from clinica.utils.stream import cprint

    clinical_specifications_folder = Path(__file__).parents[1] / "specifications"
    if not clinical_specifications_folder.exists():
        msg = (
            f"{clinical_specifications_folder} folder cannot be found ! "
            "This is an internal folder of Clinica."
        )
        cprint(msg, lvl="error")
        raise FileNotFoundError(msg)

    cprint("Creating modality agnostic files...", lvl="info")
    readme_data = {
        "link": "http://adni.loni.usc.edu/study-design/collaborative-studies/aibl/",
        "desc": (
            "The Australian Imaging, Biomarker & Lifestyle Flagship Study of Ageing (AIBL) seeks to discover which "
            "biomarkers, cognitive characteristics, and health and lifestyle factors determine the development of AD. "
            "Although AIBL and ADNI have many of the same goals, there are differences between the two projects."
        ),
    }
    write_modality_agnostic_files(
        study_name=StudyName.AIBL,
        readme_data=readme_data,
        bids_dir=output_dataset,
    )
    cprint("Creating participants.tsv...", lvl="info")
    create_participants_tsv_file(
        output_dataset,
        clinical_specifications_folder,
        input_clinical_data,
        delete_non_bids_info=True,
    )
    cprint("Creating sessions files...", lvl="info")
    create_sessions_tsv_file(
        output_dataset, input_clinical_data, clinical_specifications_folder
    )
    cprint("Creating scans files...", lvl="info")
    create_scans_tsv_file(
        output_dataset, input_clinical_data, clinical_specifications_folder
    )



"""This module contains utilities to grab or download files for Clinica."""

import hashlib
import os
from collections import namedtuple
from enum import Enum
from functools import partial
from pathlib import Path
from typing import Callable, Dict, List, Optional, Tuple, Union


class DatasetType(str, Enum):
    """Defines the possible types of datasets in Clinica."""

    RAW = "raw"
    DERIVATIVE = "derivative"


RemoteFileStructure = namedtuple("RemoteFileStructure", ["filename", "url", "checksum"])


def insensitive_glob(pattern_glob: str, recursive: Optional[bool] = False) -> List[str]:
    """This function is the glob.glob() function that is insensitive to the case.

    Parameters
    ----------
    pattern_glob : str
        Sensitive-to-the-case pattern.

    recursive : bool, optional
        Recursive parameter for `glob.glob()`.
        Default=False.

    Returns
    -------
    List[str] :
        Insensitive-to-the-case pattern.
    """
    from glob import glob

    def either(c: str) -> str:
        return "[%s%s]" % (c.lower(), c.upper()) if c.isalpha() else c

    return glob("".join(map(either, pattern_glob)), recursive=recursive)


def determine_caps_or_bids(input_dir: Union[str, os.PathLike]) -> bool:
    """Determine if the `input_dir` is a CAPS or a BIDS folder.

    Parameters
    ----------
    input_dir : os.PathLike
        The input folder.

    Returns
    -------
    bool :
        True if `input_dir` is a BIDS folder, False if `input_dir`
        is a CAPS folder or could not be determined.
    """
    input_dir = Path(input_dir)
    subjects_dir = input_dir / "subjects"
    groups_dir = input_dir / "groups"

    dir_to_look = subjects_dir if subjects_dir.is_dir() else input_dir
    subjects_sub_folders = _list_subjects_sub_folders(dir_to_look, groups_dir)
    if subjects_dir.is_dir():
        return False
    return len(subjects_sub_folders) > 0


def _list_subjects_sub_folders(
    root_dir: os.PathLike, groups_dir: os.PathLike
) -> List[os.PathLike]:
    from clinica.utils.stream import cprint

    root_dir = Path(root_dir)
    groups_dir = Path(groups_dir)
    warning_msg = (
        f"Could not determine if {groups_dir.parent} is a CAPS or BIDS directory. "
        "Clinica will assume this is a CAPS directory."
    )
    folder_content = [f for f in root_dir.iterdir()]
    subjects_sub_folders = [
        sub for sub in folder_content if (sub.name.startswith("sub-") and sub.is_dir())
    ]
    if len(subjects_sub_folders) == 0 and not groups_dir.is_dir():
        cprint(msg=warning_msg, lvl="warning")
    return subjects_sub_folders


def _validate_folder_existence(
    directory: Union[str, os.PathLike],
    folder_type: DatasetType,
) -> Path:
    """Utility function which performs checks common to BIDS and CAPS folder structures.

    Parameters
    ----------
    directory : PathLike or str
        Directory to check.

    folder_type : DatasetType
        The type of directory.

    Returns
    -------
    Path :
        The directory as a Path.
    """
    from clinica.utils.exceptions import ClinicaBIDSError, ClinicaCAPSError

    try:
        directory = Path(directory)
    except TypeError:
        raise TypeError(
            f"Argument you provided to check_{folder_type.value.lower()}_folder() is not a valid folder name."
        )

    if not directory.is_dir():
        raise (
            ClinicaBIDSError if folder_type == DatasetType.RAW else ClinicaCAPSError
        )(
            f"The {folder_type.value} directory you gave is not a folder.\n"
            "Error explanations:\n"
            f"\t- Clinica expected the following path to be a folder: {directory}\n"
            "\t- If you gave relative path, did you run Clinica on the good folder?"
        )

    return directory


_validate_bids_folder_existence = partial(
    _validate_folder_existence, folder_type=DatasetType.RAW
)
_validate_caps_folder_existence = partial(
    _validate_folder_existence, folder_type=DatasetType.DERIVATIVE
)


def check_bids_folder(bids_directory: Union[str, os.PathLike]) -> None:
    """Check if provided `bids_directory` is a BIDS folder.

    Parameters
    ----------
    bids_directory : PathLike
        The input folder to check.

    Raises
    ------
    ValueError :
        If `bids_directory` is not a string.

    ClinicaBIDSError :
        If the provided path does not exist, or is not a directory.
        If the provided path is a CAPS folder (BIDS and CAPS could
        be swapped by user). We simply check that there is not a folder
        called 'subjects' in the provided path (that exists in CAPS hierarchy).
        If the provided folder is empty.
        If the provided folder does not contain at least one directory whose
        name starts with 'sub-'.
    """
    bids_directory = _validate_bids_folder_existence(bids_directory)
    _check_dataset_description_exists_in_bids(bids_directory)
    _check_bids_is_not_caps(bids_directory)
    _check_bids_is_not_empty(bids_directory)
    _check_bids_has_at_least_one_subject_folder(bids_directory)


def _check_dataset_description_exists(directory: Path, folder_type: DatasetType):
    from clinica.utils.exceptions import ClinicaBIDSError, ClinicaCAPSError

    if not (directory / "dataset_description.json").exists():
        raise (
            ClinicaBIDSError if folder_type == DatasetType.RAW else ClinicaCAPSError
        )(
            f"The {folder_type.value} directory ({directory}) you provided is missing "
            "a dataset_description.json file."
        )


_check_dataset_description_exists_in_bids = partial(
    _check_dataset_description_exists, folder_type=DatasetType.RAW
)
_check_dataset_description_exists_in_caps = partial(
    _check_dataset_description_exists, folder_type=DatasetType.DERIVATIVE
)


def _check_bids_is_not_caps(bids_directory: Path):
    from clinica.utils.exceptions import ClinicaBIDSError

    if (bids_directory / "subjects").is_dir():
        raise ClinicaBIDSError(
            f"The BIDS directory ({bids_directory}) you provided seems to "
            "be a CAPS directory due to the presence of a 'subjects' folder."
        )


def _check_bids_is_not_empty(bids_directory: Path):
    from clinica.utils.exceptions import ClinicaBIDSError

    if (
        len(
            [
                f
                for f in bids_directory.iterdir()
                if f.name != "dataset_description.json"
            ]
        )
        == 0
    ):
        raise ClinicaBIDSError(
            f"The BIDS directory you provided is empty. ({bids_directory})."
        )


def _check_bids_has_at_least_one_subject_folder(bids_directory: Path):
    from clinica.utils.exceptions import ClinicaBIDSError

    if len([f for f in bids_directory.iterdir() if f.name.startswith("sub-")]) == 0:
        raise ClinicaBIDSError(
            "Your BIDS directory does not contains a single folder whose name "
            "starts with 'sub-'. Check that your folder follow BIDS standard."
        )


def check_caps_folder(caps_directory: Union[str, os.PathLike]) -> None:
    """Check if provided `caps_directory`is a CAPS folder.

    Parameters
    ----------
    caps_directory : os.PathLike
        The input folder to check.

    Raises
    ------
    ValueError :
        If `caps_directory` is not a string.

    ClinicaCAPSError :
        If the provided path does not exist, or is not a directory.
        If the provided path is a BIDS folder (BIDS and CAPS could be
        swapped by user). We simply check that there is not a folder
        whose name starts with 'sub-' in the provided path (that exists
        in BIDS hierarchy).

    Notes
    -----
    Keep in mind that a CAPS folder can be empty.
    """
    from clinica.utils.exceptions import ClinicaCAPSError

    caps_directory = _validate_caps_folder_existence(caps_directory)
    _check_dataset_description_exists_in_caps(caps_directory)

    sub_folders = [f for f in caps_directory.iterdir() if f.name.startswith("sub-")]
    if len(sub_folders) > 0:
        error_string = (
            "Your CAPS directory contains at least one folder whose name "
            "starts with 'sub-'. Check that you did not swap BIDS and CAPS folders.\n"
            "Folder(s) found that match(es) BIDS architecture:\n"
        )
        for directory in sub_folders:
            error_string += f"\t{directory}\n"
        error_string += (
            "A CAPS directory has a folder 'subjects' at its root, in which "
            "are stored the output of the pipeline for each subject."
        )
        raise ClinicaCAPSError(error_string)


def find_sub_ses_pattern_path(
    input_directory: os.PathLike,
    subject: str,
    session: str,
    error_encountered: list,
    results: list,
    is_bids: bool,
    pattern: str,
) -> None:
    """Appends the output path corresponding to subject, session and pattern in results.

    If an error is encountered, its corresponding message is added to the list `error_encountered`.

    Parameters
    ----------
    input_directory : str
        Path to the root of the input directory (BIDS or CAPS).

        .. warning::
            This function does not perform any check on `input_directory`.
            It is assumed that it has been previously checked by either
            `check_bids_directory` or `check_caps_directory`, and that
            the flag `is_bids` has been set accordingly.

    subject : str
        Name given to the folder of a participant (ex: sub-ADNI002S0295).

    session : str
        Name given to the folder of a session (ex: ses-M00).

    error_encountered : List
        List to which errors encountered in this function are added.

    results : List
        List to which the output path corresponding to subject, session
        and pattern is added.

    is_bids : bool
        True if `input_dir` is a BIDS folder, False if `input_dir` is a
        CAPS folder.

    pattern : str
        Define the pattern of the final file.
    """
    from clinica.utils.stream import cprint

    input_directory = Path(input_directory)
    if is_bids:
        origin_pattern = input_directory / subject / session
    else:
        origin_pattern = input_directory / "subjects" / subject / session

    current_pattern = origin_pattern / "**" / pattern
    current_glob_found = insensitive_glob(str(current_pattern), recursive=True)
    if len(current_glob_found) > 1:
        # If we have more than one file at this point, there are two possibilities:
        #   - there is a problem somewhere which made us catch too many files
        #           --> In this case, we raise an error.
        #   - we have captured multiple runs for the same subject and session
        #           --> In this case, we need to select one of these runs to proceed.
        #               Ideally, this should be done via QC but for now, we simply
        #               select the latest run and warn the user about it.
        if _are_multiple_runs(current_glob_found):
            selected = _select_run(current_glob_found)
            list_of_found_files_for_reporting = ""
            for filename in current_glob_found:
                list_of_found_files_for_reporting += f"- {filename}\n"
            cprint(
                f"More than one run were found for subject {subject} and session {session} : "
                f"\n\n{list_of_found_files_for_reporting}\n"
                f"Clinica will proceed with the latest run available, that is \n\n-{selected}.",
                lvl="warning",
            )
            results.append(selected)
        else:
            error_str = f"\t*  ({subject} | {session}): More than 1 file found:\n"
            for found_file in current_glob_found:
                error_str += f"\t\t{found_file}\n"
            error_encountered.append(error_str)
    elif len(current_glob_found) == 0:
        error_encountered.append(f"\t* ({subject} | {session}): No file found\n")
    # Otherwise the file found is added to the result
    else:
        results.append(current_glob_found[0])


def _are_multiple_runs(files: List[str]) -> bool:
    """Returns whether the files in the provided list only differ through their run number.

    The provided files must have exactly the same parent paths, extensions, and BIDS entities
    excepted for the 'run' entity which must be different.

    Parameters
    ----------
    files : List of str
        The files to analyze.

    Returns
    -------
    bool :
        True if the provided files only differ through their run number, False otherwise.
    """
    from pathlib import Path

    files = [Path(_) for _ in files]
    # Exit quickly if less than one file or if at least one file does not have the entity run
    if len(files) < 2 or any(["_run-" not in f.name for f in files]):
        return False
    try:
        _check_common_parent_path(files)
        _check_common_extension(files)
        common_suffix = _check_common_suffix(files)
    except ValueError:
        return False
    found_entities = _get_entities(files, common_suffix)
    for entity_name, entity_values in found_entities.items():
        if entity_name != "run":
            # All entities except run numbers should be the same
            if len(entity_values) != 1:
                return False
        else:
            # Run numbers should differ otherwise this is a BIDS violation at this point
            if len(entity_values) != len(files):
                return False
    return True


def _get_entities(files: List[Path], common_suffix: str) -> dict:
    """Compute a dictionary where the keys are entity names and the values
    are sets of all the corresponding entity values found while iterating over
    the provided files.

    Parameters
    ----------
    files : List of Path
        List of paths to get entities of.

    common_suffix : str
        The suffix common to all the files. This suffix will be stripped
        from the file names in order to only analyze BIDS entities.

    Returns
    -------
    dict :
        The entities dictionary.
    """
    from clinica.utils.filemanip import get_filename_no_ext

    found_entities = dict()
    for f in files:
        entities = get_filename_no_ext(f.name).rstrip(common_suffix).split("_")
        for entity in entities:
            entity_name, entity_value = entity.split("-")
            if entity_name in found_entities:
                found_entities[entity_name].add(entity_value)
            else:
                found_entities[entity_name] = {entity_value}

    return found_entities


def _check_common_properties_of_files(
    files: List[Path],
    property_name: str,
    property_extractor: Callable,
) -> str:
    """Verify that all provided files share the same property and return its value.

    Parameters
    ----------
    files : List of Paths
        List of file paths for which to verify common property.

    property_name : str
        The name of the property to verify.

    property_extractor : Callable
        The function which is responsible for the property extraction.
        It must implement the interface `property_extractor(filename: Path) -> str`

    Returns
    -------
    str :
        The value of the common property.

    Raises
    ------
    ValueError :
        If the provided files do not have the same given property.
    """
    extracted_properties = {property_extractor(f) for f in files}
    if len(extracted_properties) != 1:
        raise ValueError(
            f"The provided files do not share the same {property_name}."
            f"The following {property_name}s were found: {extracted_properties}"
        )
    return extracted_properties.pop()


def _get_parent_path(filename: Path) -> str:
    return str(filename.parent)


def _get_extension(filename: Path) -> str:
    return "".join(filename.suffixes)


def _get_suffix(filename: Path) -> str:
    from clinica.utils.filemanip import get_filename_no_ext

    return f"_{get_filename_no_ext(filename.name).split('_')[-1]}"


_check_common_parent_path = partial(
    _check_common_properties_of_files,
    property_name="parent path",
    property_extractor=_get_parent_path,
)


_check_common_extension = partial(
    _check_common_properties_of_files,
    property_name="extension",
    property_extractor=_get_extension,
)


_check_common_suffix = partial(
    _check_common_properties_of_files,
    property_name="suffix",
    property_extractor=_get_suffix,
)


def _select_run(files: List[str]) -> str:
    import numpy as np

    runs = [int(_get_run_number(f)) for f in files]
    return files[np.argmax(runs)]


def _get_run_number(filename: str) -> str:
    import re

    matches = re.match(r".*_run-(\d+).*", filename)
    if matches:
        return matches[1]
    raise ValueError(f"Filename {filename} should contain one and only one run entity.")


def _check_information(information: Dict) -> None:
    if not isinstance(information, (dict, list)):
        raise TypeError(
            "A dict or list of dicts must be provided for the argument 'information'"
        )

    if isinstance(information, list):
        for item in information:
            if not all(elem in item for elem in ["pattern", "description"]):
                raise ValueError(
                    "'information' must contain the keys 'pattern' and 'description'"
                )

            if not all(
                elem in ["pattern", "description", "needed_pipeline"]
                for elem in item.keys()
            ):
                raise ValueError(
                    "'information' can only contain the keys 'pattern', 'description' and 'needed_pipeline'"
                )

            if isinstance(item["pattern"], str) and item["pattern"][0] == "/":
                raise ValueError(
                    "pattern argument cannot start with char: / (does not work in os.path.join function). "
                    "If you want to indicate the exact name of the file, use the format "
                    "directory_name/filename.extension or filename.extension in the pattern argument."
                )
    else:
        if not all(elem in information for elem in ["pattern", "description"]):
            raise ValueError(
                "'information' must contain the keys 'pattern' and 'description'"
            )

        if not all(
            elem in ["pattern", "description", "needed_pipeline"]
            for elem in information.keys()
        ):
            raise ValueError(
                "'information' can only contain the keys 'pattern', 'description' and 'needed_pipeline'"
            )

        if isinstance(information["pattern"], str) and information["pattern"][0] == "/":
            raise ValueError(
                "pattern argument cannot start with char: / (does not work in os.path.join function). "
                "If you want to indicate the exact name of the file, use the format "
                "directory_name/filename.extension or filename.extension in the pattern argument."
            )


def _format_errors(errors: List, information: Dict) -> str:
    error_message = (
        f"Clinica encountered {len(errors)} "
        f"problem(s) while getting {information['description']}:\n"
    )
    if "needed_pipeline" in information and information["needed_pipeline"]:
        error_message += (
            "Please note that the following clinica pipeline(s) must "
            f"have run to obtain these files: {information['needed_pipeline']}\n"
        )
    error_message += "\n".join(errors)

    return error_message


def clinica_file_reader(
    subjects: List[str],
    sessions: List[str],
    input_directory: os.PathLike,
    information: Dict,
    raise_exception: Optional[bool] = True,
    n_procs: Optional[int] = 1,
):
    """Read files in BIDS or CAPS directory based on participant ID(s).

    This function grabs files relative to a subject and session list according to a glob pattern (using *)

    Parameters
    ----------
    subjects : List[str]
        List of subjects.

    sessions : List[str]
        List of sessions. Must be same size as `subjects` and must correspond.

    input_directory : PathLike
        Path to the BIDS or CAPS directory to read from.

    information : Dict
        Dictionary containing all the relevant information to look for the files.
        The possible keys are:

            - `pattern`: Required. Define the pattern of the final file.
            - `description`: Required. String to describe what the file is.
            - `needed_pipeline` : Optional. String describing the pipeline(s)
              needed to obtain the related file.

    raise_exception : bool, optional
        If True, an exception is raised if errors happen. If not, we return the file
        list as it is. Default=True.

    n_procs : int, optional
        Number of cores used to fetch files in parallel.
        If set to 1, subjects and sessions will be processed sequentially.
        Default=1.

    Returns
    -------
    results : List[str]
        List of files respecting the subject/session order provided in input.

    error_message : str
        Error message which contains all errors encountered while reading the files.

    Raises
    ------
    TypeError
        If `information` is not a dictionary.

    ValueError
        If `information` is not formatted correctly. See function `_check_information`
        for more details.
        If the length of `subjects` is different from the length of `sessions`.

    ClinicaCAPSError or ClinicaBIDSError
        If multiples files are found for 1 subject/session, or if no file is found.

        .. note::
            If `raise_exception` is False, no exception is raised.

    Notes
    -----
    This function is case-insensitive, meaning that the pattern argument can, for example,
    contain upper case letters that do not exist in the existing file path.

    You should always use `clinica_file_reader` in the following manner:

    .. code-block:: python

         try:
            file_list = clinica_file_reader(...)
         except ClinicaException as e:
            # Deal with the error

    Examples
    --------
    The paths are shortened for readability.

    You have the full name of a file.

    File `orig_nu.mgz` from FreeSurfer of subject `sub-ADNI011S4105`, session `ses-M00`
    located in mri folder of FreeSurfer output :

    >>> clinica_file_reader(
            ['sub-ADNI011S4105'],
            ['ses-M00'],
            caps_directory,
            {
                'pattern': 'freesurfer_cross_sectional/sub-*_ses-*/mri/orig_nu.mgz',
                'description': 'freesurfer file orig_nu.mgz',
                'needed_pipeline': 't1-freesurfer'
            }
        )
    ['/caps/subjects/sub-ADNI011S4105/ses-M00/t1/freesurfer_cross_sectional/sub-ADNI011S4105_ses-M00/mri/orig_nu.mgz']

    You have a partial name of the file.

    File `sub-ADNI011S4105_ses-M00_trc-18FFDG_pet.nii.gz` in BIDS directory.
    Here, filename depends on subject and session name :

    >>> clinica_file_reader(
            ['sub-ADNI011S4105'],
            ['ses-M00'],
            bids_directory,
            {
                'pattern': '*18FFDG_pet.nii*',
                'description': 'FDG PET data'
            }
        )
    ['/bids/sub-ADNI011S4105/ses-M00/pet/sub-ADNI011S4105_ses-M00_trc-18FFDG_pet.nii.gz']

    Tricky example.

    Get the file `rh.white` from FreeSurfer :

    This will fail :

    >>> clinica_file_reader(
            ['sub-ADNI011S4105'],
            ['ses-M00'],
            caps,
            {
                'pattern': 'rh.white',
                'description': 'right hemisphere of outer cortical surface.',
                'needed_pipeline': 't1-freesurfer'
            }
        )
    * More than 1 file found::
            /caps/subjects/sub-ADNI011S4105/ses-M00/t1/freesurfer_cross_sectional/fsaverage/surf/rh.white
            /caps/subjects/sub-ADNI011S4105/ses-M00/t1/freesurfer_cross_sectional/rh.EC_average/surf/rh.white
            /caps/subjects/sub-ADNI011S4105/ses-M00/t1/freesurfer_cross_sectional/sub-ADNI011S4105_ses-M00/surf/rh.white

    Correct usage (e.g. in pet-surface): pattern string must be 'sub-*_ses-*/surf/rh.white',
    or even more precise: 't1/freesurfer_cross_sectional/sub-*_ses-*/surf/rh.white'
    It then gives: ['/caps/subjects/sub-ADNI011S4105/ses-M00/t1/freesurfer_cross_sectional/sub-ADNI011S4105_ses-M00/surf/rh.white']
    """
    from clinica.utils.exceptions import ClinicaBIDSError, ClinicaCAPSError

    _check_information(information)
    pattern = information["pattern"]

    input_directory = Path(input_directory)
    is_bids = determine_caps_or_bids(input_directory)
    if is_bids:
        check_bids_folder(input_directory)
    else:
        check_caps_folder(input_directory)

    if len(subjects) != len(sessions):
        raise ValueError("Subjects and sessions must have the same length.")

    if len(subjects) == 0:
        return [], ""

    file_reader = _read_files_parallel if n_procs > 1 else _read_files_sequential
    results, errors_encountered = file_reader(
        input_directory,
        subjects,
        sessions,
        is_bids,
        pattern,
        n_procs=n_procs,
    )
    error_message = _format_errors(errors_encountered, information)

    if len(errors_encountered) > 0 and raise_exception:
        if is_bids:
            raise ClinicaBIDSError(error_message)
        else:
            raise ClinicaCAPSError(error_message)

    return results, error_message


def _read_files_parallel(
    input_directory: os.PathLike,
    subjects: List[str],
    sessions: List[str],
    is_bids: bool,
    pattern: str,
    n_procs: int,
) -> Tuple[List[str], List[str]]:
    from multiprocessing import Manager

    from joblib import Parallel, delayed

    manager = Manager()
    shared_results = manager.list()
    shared_errors_encountered = manager.list()
    Parallel(n_jobs=n_procs)(
        delayed(find_sub_ses_pattern_path)(
            input_directory,
            sub,
            ses,
            shared_errors_encountered,
            shared_results,
            is_bids,
            pattern,
        )
        for sub, ses in zip(subjects, sessions)
    )
    results = list(shared_results)
    errors_encountered = list(shared_errors_encountered)
    return results, errors_encountered


def _read_files_sequential(
    input_directory: os.PathLike,
    subjects: List[str],
    sessions: List[str],
    is_bids: bool,
    pattern: str,
    **kwargs,
) -> Tuple[List[str], List[str]]:
    errors_encountered, results = [], []
    for sub, ses in zip(subjects, sessions):
        find_sub_ses_pattern_path(
            input_directory, sub, ses, errors_encountered, results, is_bids, pattern
        )
    return results, errors_encountered


def clinica_list_of_files_reader(
    participant_ids: List[str],
    session_ids: List[str],
    bids_or_caps_directory: os.PathLike,
    list_information: List[Dict],
    raise_exception: Optional[bool] = True,
) -> List[List[str]]:
    """Read list of BIDS or CAPS files.

    This function iterates calls of `clinica_file_reader` to extract input files based
    on information given by `list_information`.

    Parameters
    ----------
    participant_ids : List[str]
        List of participant IDs.
        Example: ['sub-CLNC01', 'sub-CLNC01', 'sub-CLNC02']

    session_ids : List[str]
        List of sessions ID associated to `participant_ids`
        Example: ['ses-M00', 'ses-M18', 'ses-M00']

    bids_or_caps_directory : PathLike
        Path to the BIDS of CAPS directory to read from.

    list_information : List[Dict]
        List of information dictionaries described in `clinica_file_reader`.

    raise_exception : bool, optional
        Raise Exception or not. Defaults to True.

    Returns
    -------
    list_found_files : List[List[str]]
        List of lists of found files following order of `list_information`
    """
    from .exceptions import ClinicaBIDSError, ClinicaException

    all_errors = []
    list_found_files = []
    for info_file in list_information:
        try:
            list_found_files.append(
                clinica_file_reader(
                    participant_ids,
                    session_ids,
                    bids_or_caps_directory,
                    info_file,
                    True,
                )[0]
            )
        except ClinicaException as e:
            list_found_files.append([])
            all_errors.append(e)

    if len(all_errors) > 0 and raise_exception:
        error_message = "Clinica faced error(s) while trying to read files in your BIDS or CAPS directory.\n"
        for msg in all_errors:
            error_message += str(msg)
        raise ClinicaBIDSError(error_message)

    return list_found_files


def clinica_group_reader(
    caps_directory: os.PathLike,
    information: Dict,
    raise_exception: Optional[bool] = True,
) -> str:
    """Read files from CAPS directory based on group ID(s).

    This function grabs files relative to a group, according to a glob pattern (using *).
    Only one file can be returned, as order is arbitrary in `glob.glob()`.

    Parameters
    ----------
    caps_directory : PathLike
        Path to the input CAPS directory.

    information : Dict
        Dictionary containing all the relevant information to look for the files.
        The possible keys are:

            - `pattern`: Required. Define the pattern of the final file.
            - `description`: Required. String to describe what the file is.
            - `needed_pipeline` : Optional. String describing the pipeline(s)
              needed to obtain the related file.

    raise_exception : bool, optional
        If True, an exception is raised if errors happen.
        If not, we return the file list as it is.
        Default=True.

    Returns
    -------
    str :
        Path to the found file.

    Raises
    ------
    ClinicaCAPSError :
        If no file is found, or more than 1 files are found.
    """
    _check_information(information)
    pattern = information["pattern"]
    caps_directory = Path(caps_directory)
    check_caps_folder(caps_directory)

    current_pattern = caps_directory / "**" / pattern
    found_files = insensitive_glob(str(current_pattern), recursive=True)

    # Since we are returning found_files[0], force raising even if raise_exception is False
    # Otherwise we'll get an uninformative IndexError...
    if (len(found_files) == 0) or (len(found_files) > 1 and raise_exception is True):
        _format_and_raise_group_reader_errors(caps_directory, found_files, information)

    return found_files[0]


def _format_and_raise_group_reader_errors(
    caps_directory: os.PathLike,
    found_files: List,
    information: Dict,
) -> None:
    from clinica.utils.exceptions import ClinicaCAPSError

    error_string = (
        f"Clinica encountered a problem while getting {information['description']}. "
    )
    if len(found_files) == 0:
        error_string += "No file was found"
    else:
        error_string += f"{len(found_files)} files were found:"
        for found_file in found_files:
            error_string += f"\n\t{found_file}"
        error_string += (
            f"\n\tCAPS directory: {caps_directory}\n"
            "Please note that the following clinica pipeline(s) must have run to obtain these files: "
            f"{information['needed_pipeline']}\n"
        )
    raise ClinicaCAPSError(error_string)


def compute_sha256_hash(file_path: Path) -> str:
    """Calculate the sha256 hash of the file at path."""

    sha256hash = hashlib.sha256()
    chunk_size = 8192
    with open(file_path, "rb") as f:
        while True:
            buffer = f.read(chunk_size)
            if not buffer:
                break
            sha256hash.update(buffer)
    return sha256hash.hexdigest()


def fetch_file(
    remote: RemoteFileStructure, output_folder: Union[str, os.PathLike]
) -> Path:
    """Download a specific file and save it into the resources folder of the package.

    Parameters
    ----------
    remote : RemoteFileStructure
        Structure containing url, filename and checksum.

    output_folder : str or PathLike
        Absolute path where the file will be downloaded.
        The name of the file will be the same as the downloaded file.

    Returns
    -------
    file_path : Path
        The path to the downloaded file.
    """
    import shutil
    import ssl
    from urllib.error import URLError
    from urllib.request import Request, urlopen

    from clinica.utils.stream import cprint

    output_folder = Path(output_folder)
    if not output_folder.exists():
        cprint(
            msg=f"The path {output_folder} to store the downloaded file does not exist",
            lvl="warning",
        )
        cprint(msg="Stop Clinica and handle this error", lvl="warning")
    file_path = output_folder / remote.filename
    # Download the file from `url` and save it locally under `file_name`:
    gcontext = ssl.SSLContext(protocol=ssl.PROTOCOL_TLS_CLIENT)
    gcontext.load_default_certs()
    req = Request(remote.url + remote.filename)
    try:
        response = urlopen(req, context=gcontext)
    except URLError as e:
        if hasattr(e, "reason"):
            cprint(msg=f"We failed to reach a server. Reason: {e.reason}", lvl="error")
        elif hasattr(e, "code"):
            cprint(
                msg=f"The server could not fulfill the request. Error code: {e.code}",
                lvl="error",
            )
    else:
        try:
            with open(file_path, "wb") as out_file:
                shutil.copyfileobj(response, out_file)
        except OSError as err:
            cprint(msg="OS error: {0}".format(err), lvl="error")

    if (checksum := compute_sha256_hash(file_path)) != remote.checksum:
        raise IOError(
            f"{file_path} has an SHA256 checksum ({checksum}) from expected "
            f"({remote.checksum}), file may be corrupted."
        )
    return file_path


def get_file_from_server(
    remote_file: RemoteFileStructure,
    cache_path: Optional[str] = None,
) -> Path:
    """Download file from server.

    Parameters
    ----------
    remote_file : RemoteFileStructure
        Remote file structure defined in `clinica.utils.inputs` specifying
        the file to retrieve.

    cache_path : str, optional
        Path to cache. Default="~/.cache/clinica/data".

    Returns
    -------
    local_file : Path
        The path to the downloaded file.
    """
    from clinica.utils.stream import cprint

    if cache_path:
        cache_clinica = Path.home() / ".cache" / cache_path
    else:
        cache_clinica = Path.home() / ".cache" / "clinica" / "data"

    cache_clinica.mkdir(exist_ok=True, parents=True)
    local_file = cache_clinica / remote_file.filename

    if not local_file.exists():
        try:
            local_file = fetch_file(remote_file, cache_clinica)
        except IOError as err:
            cprint(
                msg=f"Unable to download {remote_file.filename} from {remote_file.url}: {err}",
                lvl="error",
            )

    return local_file



"""This module contains the Pipeline abstract class needed for Clinica.

Subclasses are located in clinica/pipeline/<pipeline_name>/<pipeline_name>_pipeline.py
"""

import abc
from pathlib import Path
from typing import Iterable, List, Optional, Tuple

import click
from nipype.interfaces.utility import IdentityInterface
from nipype.pipeline.engine import Node, Workflow

from clinica.utils.check_dependency import SoftwareDependency, ThirdPartySoftware
from clinica.utils.stream import log_and_warn


def clinica_pipeline(func):
    """Turns a CLI implementation into a Clinica Pipeline.

    More precisely, this decorator takes care of:
        - registering the implemented command as a subcommand
          of `clinica run`.
    """
    from clinica.pipelines.cli import cli as run_cli

    run_cli.add_command(func)

    return func


def postset(attribute, value):
    """Sets the attribute of an object after the execution.

    Args:
        attribute: An object's attribute to be set.
        value: A desired value for the object's attribute.

    Returns:
        A decorator executed after the decorated function is.
    """

    def postset_decorator(func):
        def func_wrapper(self, *args, **kwargs):
            res = func(self, *args, **kwargs)
            setattr(self, attribute, value)
            return res

        return func_wrapper

    return postset_decorator


def _detect_cross_sectional_and_longitudinal_subjects(
    subjects: Iterable[str], bids_dir: Path
) -> Tuple[List[str], List[str]]:
    """Detect whether a subject's folder has a longitudinal or cross-sectional structure.

    If it has a mixed structure, it will be considered cross-sectionnal.

    Parameters
    ----------
    subjects: Iterable of str
        List containing all the names of the subjects.

    bids_dir: Path
        The path to the bids directory.

    Returns
    -------
    cross_sectional_subjects: list of str
        List of all the subject detected with a cross_sectional organisation.

    longitudinal_subjects: list of str
        List of all the subject detected with a longitudinal organisation.

    Examples
    --------
    BIDS
    ├── sub-01
    │   └── anat
    └── sub-02
        └── ses-M000
    >>> _detect_cross_sectional_and_longitudinal_subjects(["sub-01", "sub-02"], Path("/Users/name.surname/BIDS"))
    (["sub-01"], ["sub-02])
    """
    cross_sectional_subjects = []
    longitudinal_subjects = []
    for subject in subjects:
        folders = [f.name for f in (bids_dir / subject).iterdir() if f.is_dir()]
        if not all([folder.startswith("ses-") for folder in folders]):
            cross_sectional_subjects.append(subject)
        else:
            longitudinal_subjects.append(subject)
    return cross_sectional_subjects, longitudinal_subjects


SYMBOLS = {
    "customary": ("B", "K", "M", "G", "T", "P", "E", "Z", "Y"),
    "customary_ext": (
        "byte",
        "kilo",
        "mega",
        "giga",
        "tera",
        "peta",
        "exa",
        "zetta",
        "iotta",
    ),
    "iec": ("Bi", "Ki", "Mi", "Gi", "Ti", "Pi", "Ei", "Zi", "Yi"),
    "iec_ext": (
        "byte",
        "kibi",
        "mebi",
        "gibi",
        "tebi",
        "pebi",
        "exbi",
        "zebi",
        "yobi",
    ),
}


def _bytes2human(
    n: int, format: str = "%(value).1f %(symbol)s", symbols: str = "customary"
) -> str:
    """Convert n bytes into a human readable string based on format.

    Symbols can be either "customary", "customary_ext", "iec" or "iec_ext", see: http://goo.gl/kTQMs

    License:
    Bytes-to-human / human-to-bytes converter.
    Based on: http://goo.gl/kTQMs
    Author: Giampaolo Rodola' <g.rodola [AT] gmail [DOT] com>
    License: MIT
    """
    n = int(n)
    if n < 0:
        raise ValueError("n < 0")
    symbols = SYMBOLS[symbols]
    prefix = {}
    for i, s in enumerate(symbols[1:]):
        prefix[s] = 1 << (i + 1) * 10
    for symbol in reversed(symbols[1:]):
        if n >= prefix[symbol]:
            value = float(n) / prefix[symbol]
            return format % locals()
    return format % dict(symbol=symbols[0], value=n)


def _human2bytes(s: str) -> int:
    """Attempts to guess the string format based on default symbols
    set and return the corresponding bytes as an integer.

    When unable to recognize the format ValueError is raised.

    License:
    Bytes-to-human / human-to-bytes converter.
    Based on: http://goo.gl/kTQMs
    Author: Giampaolo Rodola' <g.rodola [AT] gmail [DOT] com>
    License: MIT
    """
    init = s
    num = ""
    while s and s[0:1].isdigit() or s[0:1] == ".":
        num += s[0]
        s = s[1:]
    num = float(num)
    letter = s.strip()
    for name, sset in SYMBOLS.items():
        if letter in sset:
            break
    else:
        if letter == "k":
            # treat 'k' as an alias for 'K' as per: http://goo.gl/kTQMs
            sset = SYMBOLS["customary"]
            letter = letter.upper()
        else:
            raise ValueError("can't interpret %r" % init)
    prefix = {sset[0]: 1}
    for i, s in enumerate(sset[1:]):
        prefix[s] = 1 << (i + 1) * 10
    return int(num * prefix[letter])


def _add_session_label(filename: str) -> str:
    """Transform a cross-sectional filename into a longitudinal one.

    Examples
    --------
    >>> _add_session_label("sub-ADNI001_scans.tsv")
    sub-ADNI001_ses-M000_scans.tsv

    No modification done if filename already has a session:

    >>> _add_session_label("sub-023a_ses-M012_T1w.nii.gz")
    sub-023a_ses-M012_T1w.nii.gz

    Parameters
    ----------
    filename : str

    Returns
    -------
    str :
        filename with '_ses-M000_ added just after participant_id
    """
    import re

    m = re.search(r"(^sub-[a-zA-Z0-9]*)_(?!ses-[a-zA-Z0-9])(.*)", filename)
    try:
        return m.group(1) + "_ses-M000_" + m.group(2)
    except AttributeError:
        return filename


def _copy_and_add_session_label(src: str, dst: str) -> str:
    """Copy src to dst, but modifies the filename of the copied files if they
    match the regex template described in the _add_session_label() function.

    Parameters
    ----------
    src : str
        The path to the file that needs to be copied.

    dst : str
        The original destination for the copied file.

    Returns
    -------
    str :
        Copy with modified filename.
    """
    from shutil import copy2

    filename = Path(src).name

    return copy2(src, str(Path(dst).parent / _add_session_label(filename)))


def _convert_cross_sectional(
    bids_in: Path,
    bids_out: Path,
    cross_subjects: Iterable[str],
    long_subjects: Iterable[str],
):
    """Convert a cross-sectional-bids dataset into a longitudinal clinica-compliant dataset.

    Parameters
    ----------
    bids_in : Path
        The path to the cross-sectional bids dataset you want to convert.

    bids_out : Path
        The path to the converted longitudinal bids dataset.

    cross_subjects : Iterable of str
        Subjects in cross-sectional form (they need some adjustment).

    long_subjects : Iterable of str
        Subjects in longitudinal form (they just need to be copied).
    """
    if not bids_out.exists():
        bids_out.mkdir()
    _copy_metadata_files(bids_in, bids_out, ("dataset_description.json", ".bidsignore"))
    _convert(bids_in, bids_out, cross_subjects, cross_sectional=True)
    _convert(bids_in, bids_out, long_subjects, cross_sectional=False)


def _copy_metadata_files(bids_in: Path, bids_out: Path, files_to_copy: Iterable[str]):
    from shutil import copy2

    for file in files_to_copy:
        if (bids_in / file).is_file():
            copy2(bids_in / file, bids_out / file)


def _convert(
    bids_in: Path, bids_out: Path, subjects: Iterable[str], cross_sectional: bool
):
    for subject in subjects:
        output_folder = (
            bids_out / subject / "ses-M000" if cross_sectional else bids_out / subject
        )
        if not output_folder.exists():
            output_folder.mkdir(parents=True)
        for file_to_copy in [
            f for f in (bids_in / subject).iterdir() if not f.name.startswith(".")
        ]:
            copy_func = _copy_cross_sectional if cross_sectional else _copy_longitudinal
            copy_func(bids_in / subject / file_to_copy, output_folder)


def _copy_cross_sectional(file_to_copy: Path, output_folder: Path):
    from shutil import copy2, copytree

    if not (output_folder / file_to_copy.name).exists():
        if file_to_copy.is_dir():
            copytree(
                file_to_copy,
                output_folder / file_to_copy.name,
                copy_function=_copy_and_add_session_label,
            )
        elif file_to_copy.is_file():
            new_filename_wo_ses = _add_session_label(file_to_copy.name)
            copy2(file_to_copy, output_folder / new_filename_wo_ses)


def _copy_longitudinal(file_to_copy: Path, output_folder: Path) -> None:
    from shutil import copy2, copytree

    if not (output_folder / file_to_copy.name).exists():
        if file_to_copy.is_dir():
            copytree(file_to_copy, output_folder / file_to_copy.name)
        elif file_to_copy.is_file():
            copy2(file_to_copy, output_folder)


class Pipeline(Workflow):
    """Clinica Pipeline class.

    This class overwrites the `Workflow` to integrate and encourage the
    use of BIDS and CAPS data structures as inputs and outputs of the pipelines
    developed for the Clinica software.

    The global architecture of a Clinica pipelines is as follow:
        [ Data Input Stream ]
                |
            [ Input ]
                |
            [[ Core ]] <- Could be one or more `npe.Node`s
                |
            [ Output ]
                |
        [ Data Output Stream ]

    Attributes
    ----------
    is_built : bool
        Informs if the pipelines has been built or not.

    parameters : dict
        Parameters of the pipelines.

    info : dict
        Information presented in the associated `info.json` file.

    input_node : :obj:`npe.Node`
        Identity interface connecting inputs.

    output_node : :obj:`npe.Node`
        Identity interface connecting outputs.

    bids_directory : Path
        Directory used to read the data from, in BIDS.

    caps_directory : Path
        Directory used to read/write the data from/to in CAPS.

    subjects : list
        List of subjects defined in the `subjects.tsv` file.
        TODO(@jguillon): Check the subjects-sessions file name.

    sessions : list
        List of sessions defined in the `subjects.tsv` file.

    tsv_file : str
        Path to the subjects-sessions `.tsv` file.

    info_file : str
        Path to the associated `info.json` file.
    """

    __metaclass__ = abc.ABCMeta

    def __init__(
        self,
        bids_directory: Optional[str] = None,
        caps_directory: Optional[str] = None,
        tsv_file: Optional[str] = None,
        overwrite_caps: Optional[bool] = False,
        base_dir: Optional[str] = None,
        parameters: Optional[dict] = None,
        name: Optional[str] = None,
        ignore_dependencies: Optional[List[str]] = None,
        caps_name: Optional[str] = None,
    ):
        """Init a Pipeline object.

        Parameters
        ----------
        bids_directory : str, optional
            Path to a BIDS directory. Defaults to None.

        caps_directory : str, optional
            Path to a CAPS directory. Defaults to None.

        tsv_file : str, optional
            Path to a subjects-sessions `.tsv` file. Defaults to None.

        overwrite_caps : bool, optional
            Overwrite or not output directory. Defaults to False.

        base_dir : str, optional
            Working directory (attribute of Nipype::Workflow class). Defaults to None.

        parameters : dict, optional
            Pipeline parameters. Defaults to {}.

        name : str, optional
            Pipeline name. Defaults to None.

        ignore_dependencies : List of str
            List of names of dependencies whose installation checking procedure should be ignored.
            Defaults to None (i.e. all dependencies will be checked).

        Raises
        ------
        RuntimeError: [description]
        """
        import inspect
        import os
        from pathlib import Path
        from tempfile import mkdtemp

        from clinica.utils.caps import (
            build_caps_dataset_description,
            write_caps_dataset_description,
        )
        from clinica.utils.exceptions import ClinicaCAPSError
        from clinica.utils.inputs import check_bids_folder, check_caps_folder

        self._is_built: bool = False
        self._overwrite_caps: bool = overwrite_caps
        self._bids_directory: Optional[Path] = (
            Path(bids_directory).absolute() if bids_directory else None
        )
        self._caps_directory: Optional[Path] = (
            Path(caps_directory).absolute() if caps_directory else None
        )
        self._verbosity: str = "debug"
        self._tsv_file: Optional[Path] = Path(tsv_file).absolute() if tsv_file else None
        self._info_file: Path = (
            Path(os.path.dirname(os.path.abspath(inspect.getfile(self.__class__))))
            / "info.json"
        )
        self._info: Optional[dict] = None
        self._subjects: Optional[List[str]] = None
        self._sessions: Optional[List[str]] = None
        self._input_node = None
        self._output_node = None
        if base_dir:
            self.base_dir = Path(base_dir).absolute()
            self._base_dir_was_specified = True
        else:
            self.base_dir = Path(mkdtemp())
            self._base_dir_was_specified = False

        self._name = name or self.__class__.__name__
        self._parameters = parameters or {}
        self._dependencies: Optional[List[SoftwareDependency]] = None
        self._ignore_dependencies: List[ThirdPartySoftware] = []
        if ignore_dependencies:
            self._ignore_dependencies += [
                ThirdPartySoftware(d) for d in ignore_dependencies
            ]
        self.caps_name = caps_name

        if not self._bids_directory:
            if not self._caps_directory:
                raise RuntimeError(
                    f"The {self._name} pipeline does not contain "
                    "BIDS nor CAPS directory at the initialization."
                )
            try:
                check_caps_folder(self._caps_directory)
            except ClinicaCAPSError as e:
                desc = build_caps_dataset_description(
                    input_dir=self._caps_directory,
                    output_dir=self._caps_directory,
                    processing_name=self._name,
                    dataset_name=self.caps_name,
                    dependencies=self.dependencies,
                )
                raise ClinicaCAPSError(
                    f"{e}\nYou might want to create a 'dataset_description.json' "
                    f"file with the following content:\n{desc}"
                )
            self.is_bids_dir = False
        else:
            check_bids_folder(self._bids_directory)
            self.is_bids_dir = True
            if self._caps_directory is not None:
                if (
                    not self._caps_directory.exists()
                    or len([f for f in self._caps_directory.iterdir()]) == 0
                ):
                    self._caps_directory.mkdir(parents=True, exist_ok=True)
        if self._caps_directory:
            write_caps_dataset_description(
                input_dir=self.input_dir,
                output_dir=self._caps_directory,
                processing_name=self._name,
                dataset_name=self.caps_name,
                dependencies=self.dependencies,
            )
            check_caps_folder(self._caps_directory)
        self._compute_subjects_and_sessions()
        self._init_nodes()

    def _compute_subjects_and_sessions(self):
        from clinica.utils.participant import get_subject_session_list

        self._subjects, self._sessions = get_subject_session_list(
            self.input_dir,
            subject_session_file=self.tsv_file,
            is_bids_dir=self.is_bids_dir,
            use_session_tsv=False,
            tsv_dir=self.base_dir,
        )
        self._subjects, self._sessions = self.filter_qc()

    def filter_qc(self) -> tuple[list[str], list[str]]:
        return self._subjects, self._sessions

    @property
    def input_dir(self) -> Path:
        if self.is_bids_dir:
            return self._bids_directory
        return self._caps_directory

    @property
    def base_dir_was_specified(self) -> bool:
        return self._base_dir_was_specified

    @property
    def is_built(self) -> bool:
        return self._is_built

    @is_built.setter
    def is_built(self, value: bool):
        self._is_built = value

    @property
    def overwrite_caps(self) -> bool:
        return self._overwrite_caps

    @property
    def parameters(self) -> dict:
        return self._parameters

    @parameters.setter
    def parameters(self, value: dict):
        self._parameters = value
        # Need to rebuild input, output and core nodes
        self.is_built = False
        self.init_nodes()

    @property
    def input_node(self) -> Node:
        return self._input_node

    @property
    def output_node(self) -> Node:
        return self._output_node

    @property
    def bids_directory(self) -> Optional[Path]:
        return self._bids_directory

    @property
    def caps_directory(self) -> Optional[Path]:
        return self._caps_directory

    @property
    def subjects(self) -> List[str]:
        return self._subjects

    @subjects.setter
    def subjects(self, value: List[str]):
        self._subjects = value
        self.is_built = False

    @property
    def sessions(self) -> List[str]:
        return self._sessions

    @sessions.setter
    def sessions(self, value: List[str]):
        self._sessions = value
        self.is_built = False

    @property
    def tsv_file(self) -> Optional[Path]:
        return self._tsv_file

    @property
    def info_file(self) -> Path:
        return self._info_file

    @staticmethod
    def get_processed_images(
        caps_directory: Path, subjects: List[str], sessions: List[str]
    ) -> List[str]:
        """Extract processed image IDs in `caps_directory` based on `subjects`_`sessions`.

        Todo:
            [ ] Implement this static method in all pipelines
            [ ] Make it abstract to force overload in future pipelines
        """
        from clinica.utils.exceptions import ClinicaException
        from clinica.utils.stream import cprint

        cprint(msg="Pipeline finished with errors.", lvl="error")
        cprint(msg="CAPS outputs were not found for some image(s):", lvl="error")
        raise ClinicaException(
            "Implementation on which image(s) failed will appear soon."
        )

    def _init_nodes(self) -> None:
        """Init the basic workflow and I/O nodes necessary before build."""
        self._init_input_node()
        self._init_output_node()
        Workflow.__init__(self, self._name, self.base_dir)
        if self.input_node:
            self.add_nodes([self.input_node])
        if self.output_node:
            self.add_nodes([self.output_node])

    def _init_input_node(self) -> None:
        if self.get_input_fields():
            self._input_node = Node(
                name="Input",
                interface=IdentityInterface(
                    fields=self.get_input_fields(), mandatory_inputs=False
                ),
            )

    def _init_output_node(self) -> None:
        if self.get_output_fields():
            self._output_node = Node(
                name="Output",
                interface=IdentityInterface(
                    fields=self.get_output_fields(), mandatory_inputs=False
                ),
            )

    def has_input_connections(self) -> bool:
        """Checks if the Pipeline's input node has been connected.

        Returns
        -------
        bool :
            True if the input node is connected, False otherwise.
        """
        if self.input_node:
            return self._graph.in_degree(self.input_node) > 0
        return False

    def has_output_connections(self) -> bool:
        """Checks if the Pipeline's output node has been connected.

        Returns
        -------
        bool :
            True if the output node is connected, False otherwise.
        """
        if self.output_node:
            return self._graph.out_degree(self.output_node) > 0
        return False

    @postset("is_built", True)
    def build(self):
        """Builds the core, input and output nodes of the Pipeline.

        This method first checks it has already been run. It then checks
        the pipelines dependencies and, in this order, builds the core nodes,
        the input node and, finally, the output node of the Pipeline.

        Since this method returns the concerned object, it can be chained to
        any other method of the Pipeline class.

        Returns
        -------
        self: A Pipeline object.
        """
        if not self.is_built:
            self._check_dependencies()
            self._check_pipeline_parameters()
            if not self.has_input_connections():
                self._build_input_node()
            self._build_core_nodes()
            if not self.has_output_connections():
                self._build_output_node()
        return self

    def run(
        self,
        plugin=None,
        plugin_args=None,
        update_hash: bool = False,
        bypass_check: bool = False,
    ):
        """Executes the Pipeline.

        It overwrites the default Workflow method to check if the
        Pipeline is built before running it. If not, it builds it and then
        run it.
        It also checks whether there is enough space left on the disks, and if
        the number of threads to run in parallel is consistent with what is
        possible on the CPU.

        Args:
            Similar to those of Workflow.run.

        Returns:
            An execution graph (see Workflow.run).
        """
        import shutil

        from networkx import Graph, NetworkXError

        from clinica.utils.ux import print_failed_images

        self._handle_cross_sectional_dataset()
        if not self.is_built:
            self.build()
        if not bypass_check:
            self._check_size()
            plugin_args = self._update_parallelize_info(plugin_args)
            plugin = "MultiProc"
        exec_graph = []
        try:
            exec_graph = Workflow.run(self, plugin, plugin_args, update_hash)
            if not self.base_dir_was_specified:
                shutil.rmtree(self.base_dir)

        except RuntimeError as e:
            # Check that it is a Nipype error
            if "Workflow did not execute cleanly. Check log for details" in str(e):
                input_ids = [
                    p_id + "_" + s_id
                    for p_id, s_id in zip(self.subjects, self.sessions)
                ]
                output_ids = self.get_processed_images(
                    caps_directory=self.caps_directory,
                    subjects=self.subjects,
                    sessions=self.sessions,
                )
                missing_ids = list(set(input_ids) - set(output_ids))
                print_failed_images(self.name, missing_ids)
            else:
                raise e
        except NetworkXError:
            msg = (
                "Either all the images were already run by the pipeline "
                "or no image was found to run the pipeline."
            )
            log_and_warn(msg, UserWarning)
            exec_graph = Graph()
        return exec_graph

    @property
    def info(self) -> dict:
        import json

        if self._info is None:
            try:
                with open(self.info_file) as info_file:
                    info = json.load(info_file)
            except FileNotFoundError:
                log_and_warn(
                    f"Info file {self.info_file} for pipeline {self.name} is missing.",
                    UserWarning,
                )
                info = {}
            self._info = info
        return self._info

    @property
    def dependencies(self) -> List[SoftwareDependency]:
        if self._dependencies is None:
            dependencies = [
                SoftwareDependency.from_dict(d)
                for d in self.info.get("dependencies", [])
            ]
            dependencies = [
                d for d in dependencies if d.name not in self._ignore_dependencies
            ]
            self._dependencies = dependencies
        return self._dependencies

    def _check_dependencies(self):
        """Checks if listed dependencies are correctly installed."""
        for dependency in self.dependencies:
            dependency.check()
        self._check_custom_dependencies()

    def _check_size(self):
        """Check if the pipeline has enough space on the disk for both working directory and CAPS."""
        import sys
        from os import statvfs
        from os.path import dirname

        from clinica.utils.stream import cprint

        n_sessions = len(self.subjects)
        try:
            caps_stat = statvfs(self.caps_directory)
        except FileNotFoundError:
            # CAPS folder may not exist yet
            caps_stat = statvfs(dirname(self.caps_directory))
        wd_stat = statvfs(dirname(self.base_dir))

        # Estimate space left on partition/disk/whatever caps and wd is located
        free_space_caps = caps_stat.f_bavail * caps_stat.f_frsize
        free_space_wd = wd_stat.f_bavail * wd_stat.f_frsize

        try:
            space_needed_caps_1_session = self.info["space_caps"]
            space_needed_wd_1_session = self.info["space_wd"]
            space_needed_caps = n_sessions * _human2bytes(space_needed_caps_1_session)
            space_needed_wd = n_sessions * _human2bytes(space_needed_wd_1_session)
            error = ""
            if free_space_caps == free_space_wd:
                if space_needed_caps + space_needed_wd > free_space_wd:
                    # We assume this is the same disk
                    error = (
                        f"{error}Space needed for CAPS and working directory ("
                        f"{_bytes2human(space_needed_caps + space_needed_wd)}"
                        f") is greater than what is left on your hard drive ("
                        f"{_bytes2human(free_space_wd)})"
                    )
            else:
                if space_needed_caps > free_space_caps:
                    error = (
                        f"{error} Space needed for CAPS ("
                        f"{_bytes2human(space_needed_caps)}"
                        f") is greater than what is left on your hard drive ("
                        f"{_bytes2human(free_space_caps)})\n"
                    )
                if space_needed_wd > free_space_wd:
                    error = (
                        f"{error}Space needed for working_directory ("
                        f"{_bytes2human(space_needed_wd)}"
                        f") is greater than what is left on your hard drive ("
                        f"{_bytes2human(free_space_wd)})\n"
                    )
            if error:
                cprint(msg=f"[SpaceError] {error}", lvl="error")
                if not click.confirm(
                    "Space issues detected. Do you still want to run the pipeline?"
                ):
                    click.echo("Clinica will now exit...")
                    sys.exit()

        except KeyError:
            cprint("No info on how much size the pipeline takes. Running anyway...")

    def _update_parallelize_info(self, plugin_args: Optional[dict]) -> dict:
        """Performs some checks of the number of threads given in parameters,
        given the number of CPUs of the machine in which clinica is running.
        We force the use of plugin MultiProc

        Author: Arnaud Marcoux
        """
        from multiprocessing import cpu_count

        n_cpu = cpu_count()
        ask_user = False

        try:
            # if no --n_procs arg is used, plugin_arg is None
            # so we need a try / except block
            n_thread_cmdline = plugin_args["n_procs"]
            if n_thread_cmdline > n_cpu:
                msg = (
                    f"You are trying to run clinica with a number of threads ({n_thread_cmdline}) superior to your "
                    f"number of CPUs ({n_cpu})."
                )
                log_and_warn(msg, UserWarning)
                ask_user = True
        except TypeError:
            log_and_warn(
                (
                    f"You did not specify the number of threads to run in parallel (--n_procs argument)."
                    f"Computation time can be shorten as you have {n_cpu} CPUs on this computer. "
                    f"We recommend using {n_cpu - 1} threads."
                ),
                UserWarning,
            )
            ask_user = True
        if ask_user:
            n_procs = click.prompt(
                text="How many threads do you want to use?",
                default=max(1, n_cpu - 1),
                show_default=True,
            )
            click.echo(
                f"Number of threads set to {n_procs}. You may set the --n_procs argument "
                "to disable this message for future calls."
            )
            if plugin_args:
                plugin_args["n_procs"] = n_procs
            else:
                plugin_args = {"n_procs": n_procs}

        return plugin_args

    def _handle_cross_sectional_dataset(self):
        """Check if the dataset is longitudinal or cross-sectional.

        If it is cross-sectional, propose to convert it to a longitudinal layout.
        """
        if self.bids_directory is None:
            return
        subjects = [
            f.name
            for f in self.bids_directory.iterdir()
            if (self.bids_directory / f).is_dir() and f.name.startswith("sub-")
        ]
        (
            cross_sectional_subjects,
            longitudinal_subjects,
        ) = _detect_cross_sectional_and_longitudinal_subjects(
            subjects, self.bids_directory
        )
        if cross_sectional_subjects:
            self._convert_to_longitudinal_if_user_agrees(
                cross_sectional_subjects, longitudinal_subjects
            )

    def _convert_to_longitudinal_if_user_agrees(
        self,
        cross_sectional_subjects: List[str],
        longitudinal_subjects: List[str],
    ):
        import sys

        from clinica.utils.stream import cprint

        log_and_warn(
            (
                f"The following subjects of the input dataset {self.bids_directory} seem to "
                "have a cross-sectional layout which is not supported by Clinica:\n"
                + "\n- ".join(cross_sectional_subjects)
            ),
            UserWarning,
        )
        proposed_bids = (
            self.bids_directory.parent / f"{self.bids_directory.name}_clinica_compliant"
        )
        if not click.confirm(
            "Do you want to proceed with the conversion in another folder? "
            "(Your original BIDS folder will not be modified "
            f"and the folder {proposed_bids} will be created.)"
        ):
            click.echo(
                "Clinica does not support cross-sectional layout for BIDS dataset input. "
                "To run the pipeline, please provide a dataset in longitudinal format or accept "
                "the automatic conversion. Clinica will now exit."
            )
            sys.exit()
        cprint("Converting cross-sectional dataset into longitudinal...")
        _convert_cross_sectional(
            self.bids_directory,
            proposed_bids,
            cross_sectional_subjects,
            longitudinal_subjects,
        )
        cprint(
            f"Conversion succeeded. Your clinica-compliant dataset is located here: {proposed_bids}. "
            "The pipeline will run using this new dataset as input."
        )
        self._bids_directory = proposed_bids
        self._compute_subjects_and_sessions()

    @abc.abstractmethod
    def _build_core_nodes(self):
        """Builds the Pipeline's core nodes.

        This method should use and connect to the `Pipeline.input_node` one
        or more core `Node`s. The outputs of the core processing should then be
        connected to the `Pipeline.output_node`.
        """

    @abc.abstractmethod
    def _build_input_node(self):
        """Builds the Pipeline's input data stream node.

        Warnings:
            This method does not modify the `Pipeline.input_node` (see the
            notes about the global architecture in the class documentation).
        """

    @abc.abstractmethod
    def _build_output_node(self):
        """Builds the Pipeline's output data stream node.

        Warnings:
            This method does not modify the `Pipeline.output_node` (see the
            notes about the global architecture in the class documentation).
        """

    @abc.abstractmethod
    def get_input_fields(self) -> List[str]:
        """Lists the input fields of the Pipeline.

        Returns:
            A list of strings defining the fields of the `IdentityInterface`
            of the `Pipeline.input_node`.
        """

    @abc.abstractmethod
    def get_output_fields(self) -> List[str]:
        """Lists the output fields of the Pipeline.

        Returns:
            A list of strings defining the fields of the `IdentityInterface`
            of the `Pipeline.output_node`.
        """

    @abc.abstractmethod
    def _check_custom_dependencies(self) -> None:
        """Check dependencies provided by the developer."""

    @abc.abstractmethod
    def _check_pipeline_parameters(self) -> None:
        """Check pipeline parameters."""
